#include "capd/capdlib.h"
#include "capd/mpcapdlib.h"
#include "capd/matrixAlgorithms/capd2alglib.h"

#ifndef __CR3BP_SOLVER_H__
#define __CR3BP_SOLVER_H__

typedef capd::dynset::FactorReorganization<capd::dynset::PartialQRWithPivoting<6> >  PartialQR;
typedef capd::dynset::C1DoubletonSet<capd::IMatrix,PartialQR> PC1Rect2Set;
typedef capd::dynset::C1HOSet<PC1Rect2Set> PC1HORect2Set;
typedef capd::dynset::C2DoubletonSet<capd::IMatrix,PartialQR> PC2Rect2Set;

typedef capd::dynset::C1DoubletonSet<capd::MpIMatrix,PartialQR> MpPC1Rect2Set;
typedef capd::dynset::C1HOSet<PC1Rect2Set> MpPC1HORect2Set;
typedef capd::dynset::C2DoubletonSet<capd::MpIMatrix,PartialQR> MpPC2Rect2Set;

typedef capd::dynset::CnRect2Set<capd::IMatrix,PartialQR> PCnRect2Set;

typedef capd::ICnOdeSolver CnOdeSolver;
typedef capd::ICnPoincareMap CnPoincareMap;

/**
 * Nonrigorous solver (long double precision) of the planar system
 */ 
struct DPCR3BP{
  DPCR3BP(int order=20);
  void setMu(long double mu);

  capd::LDMap vf, exvf;
  capd::LDOdeSolver solver, exsolver;
  capd::LDCoordinateSection section, exsection;
  capd::LDPoincareMap pm, expm;


  double mu;
};

// #########################################################################

/**
 * Nonrigorous solver (long double precision) of the full 3D system
 */ 
struct DCR3BP{
  DCR3BP(int order=20);
  void setMu(long double mu);
  void setSection(int index);

  capd::LDMap vf, exvf;
  capd::LDC2OdeSolver solver, exsolver;
  capd::LDCoordinateSection y, dx, dz, exsection;
  capd::LDC2PoincareMap pm, expm;

  long double mu;
  typedef capd::LDVector Vector;
  typedef capd::LDMatrix Matrix;
  typedef capd::LDHessian Hessian;
};

// #########################################################################

/**
 * Nonrigorous solver (arbitratry precision) of the full 3D system
 */ 
struct MpCR3BP{
  MpCR3BP(int order=50, double tol=1e-60);
  void setMu(capd::MpFloat mu);

  capd::MpMap vf;
  capd::MpC2OdeSolver solver;
  capd::MpCoordinateSection section;
  capd::MpC2PoincareMap pm;
  capd::MpFloat mu;
  typedef capd::MpVector Vector;
  typedef capd::MpMatrix Matrix;
  typedef capd::MpHessian Hessian;
};

// #########################################################################
/**
 * Rigorous C^0-C^1 solver of the planar system
 */ 
struct IPCR3BP{
  IPCR3BP(int order=20, double tol=1e-18);
  void setMu(capd::interval mu);

  capd::IMap vf, exvf, Jacobi;
  capd::ICoordinateSection section, exsection;
  capd::IOdeSolver solver, exsolver;
  capd::IPoincareMap pm, expm;
};

// #########################################################################
/**
 * Rigorous C^0-C^1-C^2 solver of the full 3D system.
 * Computation are based on interval arithmetics
 * in double.
 */ 
struct ICR3BP{
  ICR3BP(int order=20, double tol=1e-18);
  void setMu(capd::interval mu);
  void setSection(int index);

  capd::IMap vf, c2vf, cnvf, exvf, exc2vf, excnvf, Jacobi;
  capd::ICoordinateSection y, dx, dz, exsection;
  capd::IOdeSolver solver, exsolver;
  capd::IC2OdeSolver c2solver, exc2solver;
  CnOdeSolver cnsolver, excnsolver;
  capd::IPoincareMap pm, expm;
  capd::IC2PoincareMap c2pm, exc2pm;
  CnPoincareMap cnpm, excnpm;
  capd::interval mu;
};

// #########################################################################
/**
 * Rigorous C^0-C^1-C^2 solver of the full 3D system.
 * Computation are based on interval arithmetics
 * in arbitrary precision.
 */ 
struct MpICR3BP{
  MpICR3BP(int order=40, double tol=1e-30);
  void setMu(capd::MpInterval mu);
  void setOrderAndToleranve(int order, double tol);

  typedef capd::MpIC2OdeSolver MpSolver;
  typedef capd::MpIC2PoincareMap MpPoincareMap;
  typedef capd::MpICnOdeSolver CnSolver;
  typedef capd::MpICnPoincareMap CnPoincareMap;

  capd::MpIMap vf, cnvf, Jacobi;
  capd::MpICoordinateSection section;
  MpSolver solver;
  MpPoincareMap pm;
  CnSolver cnsolver;
  CnPoincareMap cnpm;
  capd::MpInterval mu;
};

#endif
