#ifndef _HALO_CONTINUATION_
#define _HALO_CONTINUATION_

#include "BaseTask.h"
#include "SymBreakingBifurcationHaloAlgs.h"

using namespace capd;
using namespace std;

// #########################################################################
/**
 * This function checks if
 (1) for all x\in c there holds a \cap (b+x) \neq \emptyset.
  @param a,b,c - intervals
  @return true - if the condition (1) is validated and false otherwise.
  @remark - the function may return false even if (1) is satisfied.    
 */
bool checkOverlapingIntervals(interval a, interval b, interval c);

// #########################################################################
/**
 * This function checks if the two segments of Halo curves are overlaping.
 */ 
bool checkNonemptyIntersection(const AffineSet& s1, const AffineSet& s2, int i, int j, int c, double z);

// #########################################################################
/**
 * This function checks if segments of Halo curves returned by many tasks are overlaping.
 */ 
bool checkNonemptyIntersection(std::vector<BaseTask*>& tasks, int i, int j, int c);

// #########################################################################
/**
 * This function checks if an interval vector v is a subset of affine set s.
 */ 
bool checkInclusion(const capd::IVector& v, const AffineSet& s);

// ######################################################################
/**
 * The function is used to initialize computation of a branch of Halo orbits.
 * It checks, if it is possible to vaidate Halo orbit at selected value of continuation parameter.
 * On success, box contains a bound on this orbit.
 */  
bool proveHaloOrbitAtASinglePoint(ICR3BP& icr3bp, IVector halo, IVector& box, int i, int j, int c, int pi1, int pi2, double size);

// ######################################################################

/**
 * This task validates the existence of a branch of period-1 Halo orbits parameterized by one variable.
 * We solve u_i(u_c) and u_j(u_c), where u_i,u_j,u_c are any permutation of x,z,dy.
 * The equation to be solved reads
 *
 * \pi_{x',z'}P(x,0,z,0,dy,0) = (0,0)
 *
*/
struct HaloContinuationTask : public BaseTask  
{
  static double minParameterStep;
  static double successSetSizeFactor;
  static double successParamSizeFactor;
  static double failureParamSizeFactor;
  static bool isInitialBoxSizeMaximal;
  static bool isAdaptiveSectionChangeAllowed;
  static bool testComputation;
  
  int i,j,c,pi1,pi2;
  double sizeI, sizeJ;
  const interval SI, SJ;
  LDVector seed;
  AffineSet boundFirst, boundLast, bound;
  interval lastSuccessSizeI=0;
  interval lastSuccessSizeJ=0;
  interval si,sj;
  IVector* implicitFunctionDerivative = nullptr;
  
  HaloContinuationTask(interval param, double stepParam, int taskNo, int i, int j, int c, double si, double sj, LDVector seed);
  bool initComputation(ICR3BP& icr3bp, IVector halo, IVector& box);
  void selecSection(ICR3BP& icr3bp, DCR3BP& dcr3bp);
  void run(unsigned id);

  bool acceptEnclosure(const AffineSet& bound, double param);
  long double getSeedParameter(long double current, long double next) { return 0.5*(current+next); }

  ~HaloContinuationTask(){
    if(implicitFunctionDerivative!=nullptr)
      delete implicitFunctionDerivative;
  }
};

#endif
