/**
 * This file cotains routines which comute derivatives
 * of the reduced bifurcation function -- small 'g' in the article.
 */ 
#ifndef __LYAPUNOV_SCHMIDT_REDUCTION_H__
#define __LYAPUNOV_SCHMIDT_REDUCTION_H__

/**
 * Given derivative and normalized hessian of Poincare map
 * the function computes DY and normalized D2Y from the equation
 * P_{dx} (x,z,dy(x,z))=0
 * assuming it has unique solution.
 */ 
template<class Matrix,class Hessian>
void solveImplicitEquation(const Matrix& D, const Hessian& H, Matrix& DY, Hessian& D2Y){
  typedef typename Matrix::ScalarType Scalar;
  Scalar dx = -D[3][0]/D[3][4];
  Scalar dz = -D[3][2]/D[3][4];
  
  DY.clear();
  DY[0][0] = DY[2][2] = 1.;
  DY[4][0] = dx;
  DY[4][2] = dz;
  
  Scalar dx2 = -(H(3,0,0) + H(3,0,4)*dx + H(3,4,4)*sqr(dx))/D[3][4];  
  Scalar dz2 = -(H(3,2,2) + H(3,2,4)*dz + H(3,4,4)*sqr(dz))/D[3][4];  
  Scalar dxdz = -(H(3,0,2) + H(3,0,4)*dz + H(3,2,4)*dx + 2.*H(3,4,4)*dx*dz)/D[3][4];  

  D2Y.clear();
  D2Y(4,0,0) = dx2;
  D2Y(4,2,2) = dz2;
  D2Y(4,0,2) = dxdz;
} 

// #####################################################################

/**
 * Given third-order jet of Poincare map
 * the function computes third-order jet of dy(x,z) from the equation
 * P_{dx} (x,z,dy(x,z))=0
 * assuming it has unique solution.
 * @note - jet is assumed to contain normalized derivatives.
 */ 
template<class Jet>
void solveImplicitEquation(const Jet& D, Jet& DY){
  // here we have some redundant code
  typedef typename Jet::ScalarType Scalar;
  DY.clear();

  // first order derivatives
  Scalar dx = -D(3,0)/D(3,4);
  Scalar dz = -D(3,2)/D(3,4);  
  DY(0,0) = DY(2,2) = 1.;
  DY(4,0) = dx;
  DY(4,2) = dz;

  // second order derivatives
  Scalar dx2 = -(D(3,0,0) + dx*(D(3,0,4) + dx*D(3,4,4)))/D(3,4);  
  Scalar dz2 = -(D(3,2,2) + dz*(D(3,2,4) + dz*D(3,4,4)))/D(3,4);  
  Scalar dxdz = -(D(3,0,2) + D(3,0,4)*dz + D(3,2,4)*dx + 2.*D(3,4,4)*dx*dz)/D(3,4);  
  DY(4,0,0) = dx2;
  DY(4,2,2) = dz2;
  DY(4,0,2) = dxdz;

  // third order derivatives
  Scalar dx3 = D(3,0,0,0) + dx2*(D(3,0,4)+2.*D(3,4,4)*dx)
             + dx*( D(3,0,0,4) + dx*(D(3,0,4,4) + dx*D(3,4,4,4)) );
  DY(4,0,0,0) = -dx3/D(3,4);

  Scalar dz3 = D(3,2,2,2) + dz2*(D(3,2,4)+2.*D(3,4,4)*dz)
             + dz*( D(3,2,2,4) + dz*(D(3,2,4,4) + dz*D(3,4,4,4)) );
  DY(4,2,2,2) = -dz3/D(3,4);

  Scalar dx2dz = D(3,0,0,2) + (D(3,0,0,4)+dx*D(3,0,4,4))*dz + D(3,0,4)*dxdz
               + (D(3,0,2,4) + D(3,2,4,4)*dx + (D(3,0,4,4)+3*D(3,4,4,4)*dx)*dz + 2*D(3,4,4)*dxdz)*dx
               + (D(3,2,4)+2*D(3,4,4)*dz)*dx2;

  DY(4,0,0,2) = -dx2dz/D(3,4);

  Scalar dxdz2 = D(3,0,2,2) + (D(3,2,2,4)+dz*D(3,2,4,4))*dx + D(3,2,4)*dxdz
               + (D(3,0,2,4) + D(3,0,4,4)*dz + (D(3,2,4,4)+3*D(3,4,4,4)*dz)*dx + 2*D(3,4,4)*dxdz)*dz
               + (D(3,0,4)+2*D(3,4,4)*dx)*dz2;

  DY(4,0,2,2) = -dxdz2/D(3,4);
}

#endif
