/**
 * This file contains auxililary routines
 */ 

#ifndef __AUXIL_H__
#define __AUXIL_H__

// #########################################################################
/**
 * Auxiliary function that inverse 2x2 matrix. It is assumed the matrix is invertible.
 * Otherwise, an exception (division by zero) will be thrown.
 */
template<class M>
inline M matrixInverse(M m){
  M r(2,2);
  auto d = m(1,1)*m(2,2) - m(1,2)*m(2,1);
  r(1,1) = m(2,2)/d;
  r(2,2) = m(1,1)/d;
  r(1,2) = -m(1,2)/d;
  r(2,1) = -m(2,1)/d;
  return r;
}

// #########################################################################
/**
 * Auxiliary function that solves 2x2 linear system.
 * It is assumed that M is invertible. Otherwise an exception will be thrown.
*/
template<class M, class V>
inline V solveLinearSystem(const M& m, const V& v){
  V r(2);
  auto d = m(1,1)*m(2,2) - m(1,2)*m(2,1);
  r[0] = (v[0]*m(2,2)-v[1]*m(1,2))/d;
  r[1] = (v[1]*m(1,1)-v[0]*m(2,1))/d;
  return r;
}

// #########################################################################
/**
 * This function computes interval Krawczyk operator for 2-dim problem.
*/
template<class M, class V>
inline V intervalKrawczykOperator(const M& m, const V& v, const V& X){
  M C = matrixInverse(midMatrix(m));
  M c = C*m;
  c(1,1) -= capd::TypeTraits<typename M::ScalarType>::one();
  c(2,2) -= capd::TypeTraits<typename M::ScalarType>::one();
  return C*v + c*X;
}

// #########################################################################
/**
 * This function computes interval Newton operator for 2-dim problem.
*/
template<class M, class V>
inline V intervalNewtonOperator(const M& m, const V& v){
  V r(2);
  auto d = m(1,1)*m(2,2) - m(1,2)*m(2,1);
  r[0] = (v[0]*m(2,2)-v[1]*m(1,2))/d;
  r[1] = (v[1]*m(1,1)-v[0]*m(2,1))/d;
  return r;
}

// #########################################################################
/**
 * Several conversions between intervals with high precision end-points and intervals with double-precision end-points.
 * Conversions preserve directed roundings.
 */
 
inline capd::MpInterval tomp(capd::interval i){
  return capd::MpInterval(i.leftBound(),i.rightBound());
}

template<class I1, class I2>
void toMp(I1 b, I1 e, I2 i){
  while(b!=e) *i++ = tomp(*b++);
}

inline capd::interval tointv(capd::MpInterval a){
  return capd::interval(toDouble(a.leftBound(), capd::MpFloat::RoundDown) , toDouble(a.rightBound(), capd::MpFloat::RoundUp)); 
}

template<class I1, class I2>
void tointv(I1 b, I1 e, I2 i){
  while(b!=e) *i++ = tointv(*b++);
}
#endif
