#include <iostream>
#include "capd/capdlib.h"
#include "capd/mpcapdlib.h"
using namespace capd;
using namespace std;

/**
 * Auxiliary routine which solves linear 2-dim system.
 */ 
template<class M, class V>
inline V solveLinearSystem(const M& m, const V& v){
  V r(2);
  auto d = m(1,1)*m(2,2) - m(1,2)*m(2,1);
  r[0] = (v[0]*m(2,2)-v[1]*m(1,2))/d;
  r[1] = (v[1]*m(1,1)-v[0]*m(2,1))/d;
  return r;
}

/**
 * This namespace contains collection of nonrigorous auxiliary routines. 
 * They compute Poincare map, numerical approximation of periodic points, 
 * numerical approximation of bifurcation points in long double precision.
 */ 
namespace FalknerSkanNonrigorous{
  LDMap vf("var:x,y,z,beta;fun:y,z,beta*(y^2-1)-x*z,0;",2);
  LDC2OdeSolver solver(vf,10);
  LDCoordinateSection section(4,0);
  LDC2PoincareMap pm(solver,section);

  /** Compute k-th iterate of Poincare map and its derivative, i.e P^k(v) and DP^k(v).
   * @param k - iterate
   * @param v - a vector on the section
   * @param [out] D - computed DP^k(v)
   * @note the section is set such that both crossing directions are allowed. 
   *       Thus each periodic point is of even period.
   */  
  LDVector computePoincareMap(int k, LDVector v, LDMatrix& D){
    LDMatrix flowD(4,4), flowD1(4,4);

    LDVector y = pm(v,flowD);
    for(int j=1;j<k;++j){
      y = pm(y,flowD1);
      flowD = flowD1*flowD;
    }
    D = pm.computeDP(y,flowD);
    return y;
  }

  /** Compute k-th iterate of Poincare map and its derivative and normalized hessian (second order Taylor coefficients), 
   * i.e P^k(v) and DP^k(v), D^2P^k(v) .
   * @param k - iterate
   * @param v - a vector on the section
   * @param [out] D - computed DP^k(v)
   * @param [out] h - computed D^2P^k(v)
   * @note the section is set such that both crossing directions are allowed. 
   *       Thus each periodic point is of even period.
   */  
  LDVector computePoincareMap(int k, LDVector v, LDMatrix& D, LDHessian& H){
    LDMatrix flowD(4,4), flowD1(4,4);
    LDHessian flowH(4,4), flowH1(4,4);

    LDVector y = pm(v,flowD,flowH);
    for(int j=1;j<k;++j){
      y = pm(y,flowD1,flowH1);
      flowH = flowH1*flowD + flowD1*flowH;
      flowD = flowD1*flowD;
    }
    pm.computeDP(y,flowD,flowH,D,H);
    return y;
  }

  /**
   * This function searches for an approximate R-symmetric period-two point of the Poincare map.
   * It solves P_z(0,y,0) = 0 for y by Newton method.
   * @param u - a seed point for the Newton method
   * @return approximate R-symmetric period-two point.
   */ 
  LDVector findFixPoint(LDVector u, double tol=1e-12){
    LDMatrix D(4,4);
    LDVector y(4);
    long double delta;
    do{
      y = computePoincareMap(1,u,D);
      delta = y[2]/D[2][1];
      u[1] -= delta;
    }while(fabs(y[2])>tol and fabs(delta)>tol);
    return u;
  }

  /**
   * This function searches for an approximate bifurcation point at 1:k resonance.
   * The unknown are: 
   * - beta   - parameter of the system
   * - (0,y,0)- an initial condition
   * We solve system of equations 
   *     P_z(0,y,0,beta) = 0
   *     D_yP^k_z(0,y,0,beta) = 0
   * @param u - a seed point for the Newton method
   * @param k - period or resonance.
   * @return approximate bifurcation point in 1:k resonance.
   */ 
  LDVector findBifPoint(LDVector u, int k){
    LDVector rV(4), iV(4);
    LDMatrix D1(4,4), D(4,4), rVec(4,4), iVec(4,4);
    LDHessian H1(4,4), H(4,4);
    for(int i=0;i<30;++i){
      // compute first iterate of Poincare map and its derivative 
      LDVector y = computePoincareMap(1,u,D1,H1);
      // compute k-th iterate, its derivative and hessian
      computePoincareMap(k,u,D,H);
      // solve 2x2 linear system for Newton iteration
      LDVector f({y[2],D[2][1]});
      LDMatrix Df({{D1[2][1],D1[2][3]},{H(2,1,1),H(2,1,3)}});
      LDVector r = solveLinearSystem(Df,f);
      u[1] -= r[0];
      u[3] -= r[1];
    }
    return u;
  }

  void findBifPoints(){
    cout << "k=2: " << findBifPoint(LDVector({0,0.939848421803516676377,0,340}),2) << endl;
    cout << "k=3: " << findBifPoint(LDVector({0,0.939848421803516676377,0,453}),3) << endl;
    cout << "k=4: " << findBifPoint(LDVector({0,0.939904506713693437513,0,680}),4) << endl;
  }
} // namespace FalknerSkanNonrigorous

/// Global variables - ODE solvers and data structures used in the computation.

IMap vfn("var:x,y,z,beta;fun:y,z,beta*(y^2-1)-x*z,0;",3);
IMap vf("var:x,y,z,beta;fun:y,z,beta*(y^2-1)-x*z,0;",3);
ICnOdeSolver solvern(vf,8);
IC2OdeSolver solver(vf,15);
const int secCoordinate = 2;
const int bifCoordinate = 0 ;
ICoordinateSection section(4,secCoordinate);
ICnPoincareMap pmn(solvern,section);
IC2PoincareMap pm(solver,section);
IMatrix DP(4,4), D(4,4);
IHessian DH(4,4), H(4,4);

/**
 * Rigorous algorithm. It checks, of there exists a branch
 * of R-symmetric period-two points (0,y(beta),0) parametrized by the system parameter 
 *    beta \in u0[3] + box[3] =: J
 * such that 
 *    y(beta) \in u0[1] + box[1] 
 * @param [in] u0 - an approximate period-two point
 * @param [in] box - a box around u0
 * @param [out] dy_dbeta - bound on the derivative y'(beta) for all beta \in J
 * @param [out] ptrN - a place to store computed interval Newton operator
 * @param [in] n - technical parameter: number of subdivisions when compute value at center for interval Newton operator
 */ 
bool validateCurveOfFixedPoints(IVector u0, IVector box, interval& dy_dbeta, const int n = 400, interval* ptrN=nullptr){
  cout << "paramRange: " << u0[3] + box[3] << endl;
  // Compute image at centre but for all parameters beta, i.e. y0=u0[1] and beta=u0[3]+box[3].
  C0HORect2Set s0(u0);
  interval y = pm(s0)[bifCoordinate]; // Compute \pi_z P(u0)
  // subdivide u0[3]+box[3] and take interval hull of results
  IVector u = u0;
  u[3] += box[3];
  IVector l = leftVector(u), r = rightVector(u), d = (r-l)/n;
  for(int i=0;i<n;++i){
    C0HORect2Set s0(l + interval(i,i+1)*d);
    y = intervalHull(pm(s0)[bifCoordinate],y);
  }
  // Now y contains bound on \pi_z P(u0 + (0,0,0,box[3]))
  
  // Compute derivative DP(u0+box)
  C1HORect2Set s(u0+box);
  IVector v = pm(s,DP);
  DP = pm.computeDP(v,DP);
  // Select D_yP_z
  interval dy = DP[bifCoordinate][1];

  // Bound on derivative of implicit curve, provided it exists
  dy_dbeta = -DP[bifCoordinate][3]/DP[bifCoordinate][1];

  // Now we validate the existence of the curve of period-two points by means of Interval Newton Operator
  // We have to check inclusion -y/dy \subset box[1]
  interval N = -y/dy;
  cout << "y=" << y << endl;
  cout << "dy=" << dy << endl;
  cout << "N=" << N << endl;
  cout << "Y=" << box[1] << endl;
  if(ptrN)
    *ptrN = N;
  // check inclusion and return the result
  return subsetInterior(N,box[1]);
}

/**
 * This function validates if there exists a branch od zeroes of reduced bifurcation function g_k - see the article.
 * It checked, that there is a smooth function beta(y) solving  
 *    g_k(beta(y),y)=0
 * for y \in u[1] + box[1] such that
 *    beta(y) \in u0[3] + box[3]
 * @remark high precision interval arithmetics is used
 * @param [in] u0 - an approximate bifurcation point
 * @param [in] box - a box around u0 in which we validate the existence of bifurcation
 * @param [in] k - period of bifurcating branch
 * @param [in] dy_dbeta - bound on derivative of the curve of fixed points y'(beta) computed in the function "validateCurveOfFixedPoints".
 * @returns true if all required inclusions are satisfied, false otherwise.
 */ 
bool validateCurveOfPeriodKPoints(IVector u0, int k, IVector box, interval dy_dbeta){
  // The reduced bifurcation function can be bounded by g_k(beta,y) \subset \int DyP_z(t(y-y(beta))+y(beta),beta)dt
  // where y(beta) is the curve of fixed points.
  // We solve for g_k(beta(y),y)=0
  
  // First we compute value at the centre; need to use high precision for validation of period quadrupling
  // Define ODE solvers with 250 bits of mantissa and of order 50.
  MpFloat::setDefaultPrecision(250);
  MpIMap vf3d("par:beta;var:x,y,z;fun:y,z,beta*(y^2-1)-x*z;");
  MpIOdeSolver solver3d(vf3d,50);
  MpICoordinateSection section3d(3,secCoordinate);
  MpIPoincareMap pm3d(solver3d,section3d);
  vf3d.setParameter(0,MpInterval(MpFloat(u0[3].leftBound())));
  MpIVector mpu0({MpInterval(u0[0].leftBound()),MpInterval(u0[1].leftBound()),MpInterval(u0[2].leftBound())});
  MpC1Rect2Set s0(mpu0);
  MpIMatrix DP3(3,3);

  // compute derivative of DP_k(u0)
  MpIVector v3 = pm3d(s0,DP3,k);
  // and select partial derivative we need
  MpInterval y0 = pm3d.computeDP(v3,DP3)[bifCoordinate][1];
  // wrap the result into regular double-precision interval
  interval y = interval(toDouble(y0.leftBound(),MpFloat::RoundDown),toDouble(y0.rightBound(),MpFloat::RoundUp));

  // Remaining computation are performed with double-precision interval arithmetic.
  // Compute derivative and hessian of P(u0+box)
  C2Rect2Set s(u0+box);
  IVector v = pm(s,DP,DH,k);
  pm.computeDP(v,DP,DH,D,H);

  // Use mean value form to bound:
  // y = g_k(u0[1]+box[1],u0[3]) \subset g_k(u0[1],u0[3]) + D_yg_k(u0[1]+box[1],u0[3])*box[1]
  y += H(bifCoordinate,1,1)*box[1];
  // derivative of g_k wrt parameter beta along fixed point curve
  interval dy = H(bifCoordinate,1,3) + H(bifCoordinate,1,1)*dy_dbeta;

  // Interval Newton Operator in 1-dim case
  interval N = -y/dy;
  cout << "y=" << y << endl;
  cout << "dy=" << dy << endl;
  cout << "diam(y)=" << diam(y)  << endl;
  cout << "diam(dy)=" << diam(dy) << endl;
  cout << "N=" << N << endl;
  cout << "C=" << box[3] << endl;
  // check inclusion
  return subsetInterior(N,box[3]);
}

/**
 * A general scheme for proving touch-and-go bifurcation (k primes only).
 * @param u0 - an approximate bifurcation point of the form (0,y,0,beta)
 * @param box - a box centered at u0 in which we validate bifurcation
 * @param k - resonance 1:k. Assumed to be prime number (we do not check principal period).
 */ 
void proveTAG(IVector u0, IVector box, int k){
  cout << "########## proveTAG #########\n";
  // Call two general rountines which prove the existence of two intersecting curves 
  // which correspond to period-two and period-2k points.  
  interval dy_dbeta;
  cout << "FP: " << validateCurveOfFixedPoints(u0,box,dy_dbeta) << endl;
  cout << "BC: " << validateCurveOfPeriodKPoints(u0,k,box,dy_dbeta) << endl;

  // check C4 from the article
  cout << "D2G/dvdy: " << H(bifCoordinate,1,3) << endl;
  cout << "D2G/dydy: " << 2.*H(bifCoordinate,1,1) << endl;
  cout << "y'(v): " << dy_dbeta << endl;
  interval C4 = H(bifCoordinate,1,3) + 2.*H(bifCoordinate,1,1)*dy_dbeta;
  cout << "C4: " << C4 << ", " << !isSingular(C4) << endl;

  // check monotonicity, i.e. D2G/dp^2 nonzero
  cout << "Monotonicity: " << 2.*H(bifCoordinate,1,1) << ", " << !isSingular(H(bifCoordinate,1,1)) << endl << endl;
}

/**
 * A general scheme for proving period-tupling bifurcation.
 * @param u0 - an approximate bifurcation point of the form (0,y,0,beta)
 * @param box - a box centered at u0 in which we validate bifurcation
 * @param k - resonance 1:k (k=2 or k=4 only). 
 */ 
void proveTUP(IVector u0, IVector box, int k){
  cout << "########## proveTUP #########\n";
  // Call two general rountines which prove the existence of two intersecting curves 
  // which correspond to period-two and period-2k points.  
  interval dy_dbeta;
  cout << "FP: " << validateCurveOfFixedPoints(u0,box,dy_dbeta) << endl << endl;
  cout << "BC: " << validateCurveOfPeriodKPoints(u0,k,box,dy_dbeta) << endl << endl;

  // check C4 from the article
  cout << "D2G/dvdy: " << H(bifCoordinate,1,3) << endl;
  cout << "D2G/dydy: " << 2.*H(bifCoordinate,1,1) << endl;
  cout << "y'(v): " << dy_dbeta << endl;
  interval C4 = H(bifCoordinate,1,3) + 2.*H(bifCoordinate,1,1)*dy_dbeta;
  cout << "C4: " << C4 << ", " << !isSingular(C4) << endl;

  // check convexity of g_k. We need third order derivatives of Poincare map.
  IJet J(4,4,3);
  CnRect2Set s(u0+box,3);
  pmn(s,J,k);
  J = pmn.computeDP(J);
  interval C = J(bifCoordinate,1,1,1) / H(bifCoordinate,1,3);
  cout << "D3G:" << 6.*J(bifCoordinate,1,1,1) << endl;
  cout << "Convexity: " << C << ", " << !isSingular(C) << endl << endl;
  // If k=4, check if this is really quadrupling, i.e. g_2 should be nonzero
  if(k==4){
    C1Rect2Set s(u0+box);
    IMatrix DP(3,3);
    IVector tmp = pmn(s,DP,2);
    interval y = pm.computeDP(tmp,DP)[bifCoordinate][1];
    cout << "Minimal period for quadrupling: g_2=" << y << endl; 
  }
}

/**
 * This function proves that there is a branch of R-symmetric period-two points 
 * for the Falkner Skan equation.
 */ 
void validateBranchOfPeriodTwoPoints(){
  // initial seed point for the Newton method. Then it will be refined by continuation.
  LDVector u({0.,0.94,0.,2.}); 

  // parameter range [L,R]=[9/8,100000] and initial step "s" in parameter range.
  double L=1.125, R=100000, s = 1./1024;

  // For a subinterval of parameters beta\in J we will validate that
  //     y(beta) \subset y0 + box[1], 
  // where y0 is a seed value
  IVector box(4); box[1] = 1e-4*interval(-1,1);

  // Main loop - adaptive subdivision of parameter range.
  bool first = true;
  // To store bounds computed in the previous step. 
  // We need this to check, thath the bounds are such that the curve is continuous. 
  interval prevY, prevDeltaY, N;

  // counts number of subdivisions.
  int counter = 0;
  while(L<R){
    double nextL = capd::min(R,L+s); // does not need to be rigorous
    // we will try to validate branch in the parameter range [L,nextL]
    interval beta = interval(L,nextL);
    interval beta0, dydc;
    beta.split(beta0,box[3]);
    u[3] = beta0.leftBound();
    // find approximate period-two point for beta = (L+nextL)/2
    u = FalknerSkanNonrigorous::findFixPoint(u);
    IVector u0(4);
    std::copy(u.begin(),u.end(),u0.begin());
    try{
      // call general rountine which proves branch of periodic points
      if(validateCurveOfFixedPoints(u0,box,dydc,1,&N)){
        // if success, increase counter and check, 
        // that the bounds from the current and previous steps overlap in such a way,
        // that the curve must be smooth
        counter++;
        if(first){
          first = false;
        } else {
          // We need to check, that u0[1]+box[1] overlaps prevY+prevDeltaY 
          // or equivalently, that box[1] overlaps (u0[1]-prevY) + prevDeltaY.
          // Given that both box[1] and prevDeltaY are centered at zero,
          // it suffices to check that u0[1]-prevY belongs to prevDeltaY.
          // Then (u0[1]-prevY) + prevDeltaY contains zero as well as box[1].
          if(!subset(u0[1]-prevY,prevDeltaY)){
            cout << "Fatal error. Cannot prove that the branch of periodic points is continuous.";
            return;
          }
        }
        // accept bounds, i.e. store them in prev*
        prevY = u0[1];
        prevDeltaY = box[1];
        L = nextL;
        
        // predict parameters of Newton operator for the next step
        box[1] = interval(-1.1,1.1)*N;
        s *= 1.02;
      } else {
        // if could not check branhc of period-two points, refine parameter range (shorten it by the factor 1.2)
        // and try again
        s /= 1.2;
        cout << "Could not validate branch of period-two point for the parameter range.\nNew step in parameter range s=" << s << endl;
        if(!first)
          box[1] = prevDeltaY;
      }
    }catch(...){
      // An exception may occur when computing Poincare map (for instance the solver may be not able to compute trajectory due to blow-up)
      // In such a case, shorten the parameter range and try again.
      s /= 1.2;
      cout << "Could not validate branch of period-two point for the parameter range.\nNew step in parameter range s=" << s << endl;
      if(!first)
        box[1] = prevDeltaY;
    }
  }
  // If the main loop completes, whole range of parameters has been examined.
  cout << "Number of subintervals = " << counter << endl << endl;
}

int main(){
  cout.precision(21);
  cout << boolalpha;

  // First we prove the existence of a branch of period-two points
  validateBranchOfPeriodTwoPoints();

  // Then we prove bifurcations related to strong resonances.
  // Approximate bifurcation points found by findBifPoints
  IVector u2({0,0.939792756949623004649,0,340.18753498914353231}),
          u3({0,0.939848585715971576498,0,453.442586821384637563}),
          u4({0,0.939904499164097608161,0,679.95415507296894192});
  
  IVector box(4);
  box[1] = 1e-13*interval(-1,1); box[3] = 1e-9*interval(-1,1);
  proveTUP(u2,box,2);
  proveTAG(u3,box,3);
  proveTUP(u4,box,4);
}
