#include <iostream>
#include <fstream>
#include <atomic>
#define LOG_VALUE(x) { out << (#x) << "=" << (x) << endl; }

#include "main.h"
#include "CR3BPSolverPool.h"
#include "tictac.h"
#include "capd/normalForms/linearSubstitution.hpp"

#define CHECK_AND_LOG(m,x) { \
  bool t = x;\
  out << m << (#x) << ": " << boolalpha << t << "\n############################\n\n" << flush;\
  if(!t) this->result = false; } 


using namespace std;
using namespace capd;

// global constants

constexpr int mantissaBits = 250;

// #########################################################################
/**
 * Nonrigorous routine. It computes n-th iteation of Poincare map along with its derivative and hessian
 */ 
template<class CR3BP>
typename CR3BP::Vector computePoincareMap(CR3BP& cr3bp, int n, typename CR3BP::Vector v, typename CR3BP::Matrix& D, typename CR3BP::Hessian& H){
  typename CR3BP::Matrix flowD(6,6), flowD1(6,6);
  typename CR3BP::Hessian flowH(6,6), flowH1(6,6);

  typename CR3BP::Vector y = cr3bp.pm(v,flowD,flowH);
  for(int j=1;j<n;++j){
    y = cr3bp.pm(y,flowD1,flowH1);
    flowH = flowH1*flowD + flowD1*flowH;
    flowD = flowD1*flowD;
  }
  cr3bp.pm.computeDP(y,flowD,flowH,D,H);
  return y;
}

// #########################################################################

template<class CR3BP>
typename CR3BP::Vector findBifurcationPoint(CR3BP& cr3bp, int n, typename CR3BP::Vector v, double tolerance = 1e-12, int i=0, int j=4, int c=2, int pi1=3, int pi2=5){
  typedef typename CR3BP::Hessian Hessian;
  typedef typename CR3BP::Matrix Matrix;
  typedef typename CR3BP::Vector Vector;
  
  MESSAGE("==============> findBifPoint started");
  for(int k=0;k<100;++k){
    Matrix D(6,6), D1(6,6), DY(6,6);
    Hessian H(6,6), H1(6,6), D2Y(6,6);
    Vector y = computePoincareMap(cr3bp,1,v,D1,H1);
    Vector Dz({D1[pi1][c],D1[pi2][c]});
    Matrix Dxy({{D1[pi1][i],D1[pi1][j]},{D1[pi2][i],D1[pi2][j]}});
    Vector Dfp = -solveLinearSystem(Dxy,Dz);

    computePoincareMap(cr3bp,n,v,D,H);
    solveImplicitEquation(D,H,DY,D2Y);
    H = H*DY + D*D2Y;
    D = D*DY;
    Vector func_value({D[pi2][i],y[pi1],y[pi2]}); 
    Matrix M({
        {H(pi2,i,i),H(pi2,i,c)+H(pi2,i,i)*Dfp[0],0.},
        {D1[pi1][i],D1[pi1][c],D1[pi1][j]},
        {D1[pi2][i],D1[pi2][c],D1[pi2][j]}
      });
    PRINT(func_value);
    if(fabs(toDouble(func_value[0]))<tolerance and fabs(toDouble(func_value[1]))<tolerance and fabs(toDouble(func_value[2]))<tolerance) break;
    Vector correction = -matrixAlgorithms::gauss(M,func_value);
    PRINT(correction);
    v[i] += correction[0];
    v[c] += correction[1];
    v[j] += correction[2];
  }

  PRINT(v);
  MESSAGE("==============> findBifPoint finished");
  return v;

}

// #########################################################################
/**
 * Nonrigorous routine. It finds 'good' coordinate system for n-tupling bifurcation in the hyperplane (x,z,dy)
 * The coordinates (y,dx,dz) are untouched.
 * First axes is tangent to fixed point curve at bifurcation point.
 * Second is tangent to bifurcation curve
 * Third is orthogonal to the remaining two axes
 */ 
template<class CR3BP>
typename CR3BP::Matrix findBifurcationCoordintes(CR3BP& cr3bp, int n, typename CR3BP::Vector v){
  typedef typename CR3BP::Hessian Hessian;
  typedef typename CR3BP::Matrix Matrix;
  typedef typename CR3BP::Vector Vector;

  Matrix result = Matrix::Identity(6);
  Matrix D(6,6), DY(6,6);
  Hessian H(6,6), D2Y(6,6);
  
  // First we compute tangent vector to fixed points curve (x(z),z,dy(z))
  // and set it as the third column of the result.
  Vector y = computePoincareMap(cr3bp,1,v,D,H);
  Vector Dz({D[3][2],D[5][2]});
  Matrix Dxy({{D[3][0],D[3][4]},{D[5][0],D[5][4]}});
  Vector u = -solveLinearSystem(Dxy,Dz);
  result[0][2] = u[0];
  result[2][2] = 1.;
  result[4][2] = u[1];
  result.column(2).normalize();

  // Now we have to compute tangent vector to the bifurcation curve.
  // It can be parameterized by x.
  computePoincareMap(cr3bp,n,v,D,H);
  solveImplicitEquation(D,H,DY,D2Y);
  H = H*DY + D*D2Y;
  D = D*DY;
  result[0][0] = 1.;
  result[2][0] = -H(5,0,0)/(H(5,0,2) + H(5,0,0)*u[0]);
  result[4][0] = DY[4][0] + DY[4][2]*result[2][0];
  result.column(0).normalize();

  // The last direction is taken as vector product of remaining two vectors
  // Note, we are spanning 3D subspace
  auto p = result.column(0);
  auto q = result.column(2);
  result[0][4] = p[2]*q[4] - p[4]*q[2];
  result[2][4] = p[4]*q[0] - p[0]*q[4];
  result[4][4] = p[0]*q[2] - p[2]*q[0];
  result.column(4).normalize();

  return result;
}

// #########################################################################

mutex consoleGuard;
struct BifurcationTask {
  BifurcationTask(
    double mu, const char* msg, int period, LDVector seed,
    LDVector d=1e-17*LDVector({1.,0.,1.,0.,1.,0.}),
    int order = 40, double tol = 1e-25
  ) : result(true), isPrecomputed(false),
      mu(mu), u(seed), d(d), order(order), n(period), tolerance(tol),
      dcr3bp(order+10,1e-15*tol),cr3bpCentreSolver(order+10,1e-10*tol), cr3bpFixPointsSolver(order+10,1e-10*tol), cr3bpBifBranchSolver(order,tol),
      Dfp(2), fp(6), x(6), r(6), rBif(6),
      y0(period,MpIVector(6)),
      M(6,6), invM(6,6),
      D(period,MpIMatrix(6,6)), D0(period,MpIMatrix(6,6)), DBif(period,MpIMatrix(6,6)), DY(period,MpIMatrix(6,6)), 
      H0(period,MpIHessian(6,6)), H(period,MpIHessian(6,6)), HBif(period,MpIHessian(6,6)), D2Y(period,MpIHessian(6,6))
  {
    strcpy(this->msg,msg);
    MpFloat mass;
    mass = mu;
    dcr3bp.setMu(mass);
    cr3bpCentreSolver.setMu(MpInterval(mass));
    cr3bpFixPointsSolver.setMu(MpInterval(mass));
    cr3bpBifBranchSolver.setMu(MpInterval(mass));
    out.precision(30);
    out << boolalpha;
  }

  BifurcationTask(const char* filename) : result(true), isPrecomputed(true),
      dcr3bp(60,1e-50),cr3bpCentreSolver(40,1e-25), cr3bpFixPointsSolver(40,1e-25), cr3bpBifBranchSolver(40,1e-25),
      Dfp(2), fp(6), x(6), r(6), rBif(6),
      y0(4,MpIVector(6)),
      M(6,6), invM(6,6),
      D(4,MpIMatrix(6,6)), D0(4,MpIMatrix(6,6)), DBif(4,MpIMatrix(6,6)), DY(4,MpIMatrix(6,6)),
      H0(4,MpIHessian(6,6)), H(4,MpIHessian(6,6)), HBif(4,MpIHessian(6,6)), D2Y(4,MpIHessian(6,6))
  {
    char tmp[64];
    strcpy(msg,filename);
    ifstream file(filename);
    file.precision(50);
    file.getline(tmp,64);
    file >> n;
    file >> order;
    file >> tolerance;
    file >> d;
    file >> mu;
    file >> v;
    file >> A;
    file >> checkConvexity;
    file.close();

    MpFloat mass = mu;
    dcr3bp.setMu(mass);
    cr3bpCentreSolver.setMu(MpInterval(mass));
    cr3bpFixPointsSolver.setMu(MpInterval(mass));
    cr3bpBifBranchSolver.setMu(MpInterval(mass));
    cr3bpCentreSolver.setOrderAndToleranve(order+10,1e-10*tolerance);
    cr3bpFixPointsSolver.setOrderAndToleranve(order+10,1e-10*tolerance);
    cr3bpBifBranchSolver.setOrderAndToleranve(order,tolerance);
    out.precision(50);
    out << boolalpha;
  }

  void run();
  bool computeLyapunovSchmidtReduction(int i);
  MpInterval computeG(int i);
  MpInterval computeDxG(int i);
  bool checkBifurcationBranch();
  void recomputeC2C3AtBifPoint();

  void computeValueAtCentre(){
    try{
      MpFloat::setDefaultPrecision(mantissaBits);
      MpC1Rect2Set c(MpC1Rect2Set::C0BaseSet(this->x,this->M,MpIVector(6)),MpC1Rect2Set::C1BaseSet(this->M));
      for(int i=0;i<this->n;++i){
        this->y0[i] = this->cr3bpCentreSolver.pm(c,this->D0[i]);
        this->D0[i] = this->cr3bpCentreSolver.pm.computeDP(this->y0[i],this->D0[i]);
      }
    }catch(exception& e){
      out << e.what() << endl;
      this->result = false;
    }
  }

  void proveBranchOfFixedPoints(){
    try{
      MpFloat::setDefaultPrecision(mantissaBits);
      MpC1Rect2Set c(MpC1Rect2Set::C0BaseSet(this->x,this->M,MpIVector(6)),MpC1Rect2Set::C1BaseSet(this->M));
      MpC2Rect2Set s(MpC2Rect2Set::C0BaseSet(this->x,this->M,this->r),MpC2Rect2Set::C1BaseSet(this->M));
      CHECK_AND_LOG("Validate the existence of a branch of period-1 points:\n", proveBranchHaloOrbits(this->cr3bpFixPointsSolver.pm,this->cr3bpFixPointsSolver.pm,c,s,this->fp,0,4,2,3,5,&(this->Dfp)));
    }catch(exception& e){
      out << e.what() << endl;
      this->result = false;
    }
  }

  void evalC2(MpICR3BP& cr3bp, MpIVector _r, std::vector<MpIMatrix>& _D, std::vector<MpIHessian>& _H){
    try{
      MpFloat::setDefaultPrecision(mantissaBits);
      MpC2Rect2Set s(MpC2Rect2Set::C0BaseSet(this->x,this->M,_r),MpC2Rect2Set::C1BaseSet(this->M));
      MpIMatrix flowD(6,6);
      MpIHessian flowH(6,6);
      for(int i=0;i<this->n;++i){
        MpIVector y = cr3bp.pm(s,flowD,flowH);
        cr3bp.pm.computeDP(y,flowD,flowH,_D[i],_H[i]);
      }
    }catch(exception& e){
      out << e.what() << endl;
      this->result = false;
    }
  }

  void evalC3(){
    if(checkConvexity){ // C^3 computation only if the corresponding flag in config file is set to 1
      try{
        MpFloat::setDefaultPrecision(mantissaBits);
        MpCnRect2Set s(this->x,this->M,this->r,3);
        MpIJet J1(6,6,3);
        cr3bpBifBranchSolver.cnpm.setOrder(10);
        cr3bpBifBranchSolver.cnpm(s,J1,2);
        if(this->n==2){
          this->J = cr3bpBifBranchSolver.cnpm.computeDP(J1);
        } else { // case this->n==4
          substitutionPowerSeries(J1,J1,this->J,false);
          this->J = cr3bpBifBranchSolver.cnpm.computeDP(this->J);
        }
      }catch(exception& e){
        out << e.what() << endl;
        this->result = false;
      }
    }
  }

  void writeReport(){
    // stop clock and report time of computation
    tac(this->t,this->out);

    lock_guard<mutex> lock(consoleGuard);
    string s = out.str();
    ostringstream stream;
    stream << "./log/" << basename(msg) << "_log";
    ofstream file(stream.str());
    file << s;
    file.close();  
    cout << s << endl;
  }

  void init(){
    if(!isPrecomputed){
      v = MpVector({u[0],u[1],u[2],u[3],u[4],u[5]});
    }
    v = findBifurcationPoint(dcr3bp,n,v,1e-60);
    A =  findBifurcationCoordintes(dcr3bp,n,v);

// Create 'interval objects' from approximate vectors and matrices.
    x = vectalg::convertObject<MpIVector>(v);
    M = vectalg::convertObject<MpIMatrix>(A);
    invM = matrixAlgorithms::krawczykInverse(M);
    r = vectalg::convertObject<MpIVector>(d)*MpInterval(-1,1);
    fp = r;
  }
  
  atomic_bool result;
  bool isPrecomputed, checkConvexity;
  double mu;
  char msg[64];
  LDVector u, d;
  MpVector v;
  MpMatrix A;
  int order, n;
  double tolerance;
  MpCR3BP dcr3bp;
  MpICR3BP cr3bpCentreSolver, cr3bpFixPointsSolver, cr3bpBifBranchSolver;
  
  MpIVector Dfp, fp, x, r, rBif;
  std::vector<MpIVector> y0;
  MpIMatrix M, invM;
  std::vector<MpIMatrix> D, D0, DBif, DY;
  std::vector<MpIHessian> H0, H, HBif, D2Y;
  MpIJet J = MpIJet(6,6,3);
  ostringstream out;
  std::chrono::time_point<std::chrono::system_clock> t;
};

// #########################################################################
/**
 * This is a rigorous routine. It computes Lyapunov-Schmidt reduction, i.e.
 * it solves dy(x,z) from equation P_{dx}(x,0,z,0,dy(x,z),0) = 0
 * @result true if the existence of implicit function is validated
 * @result if success and i is equal to the period of bifurcating branch,
 *         then r[4] is modified and it will containn bound on dy(x,z) 
 */
bool BifurcationTask::computeLyapunovSchmidtReduction(int i){
// Step 1. First we compute value and derivative at (x0,z0,dy0)  
  MpInterval dy0 = y0[i][3] + D0[i][3][0]*r[0] + D0[i][3][2]*r[2];
  MpInterval Ddy = D0[i][3][4];
  
// Step 2. Add error bound from second order derivatives computed on entire set
  dy0 += H[i](3,0,0)*sqr(r[0]) + H[i](3,0,2)*r[0]*r[2] + H[i](3,2,2)*sqr(r[2]);  
  Ddy += H[i](3,4,0)*r[0] + H[i](3,4,2)*r[2] + 2.*H[i](3,4,4)*r[4];
  LOG_VALUE(y0[i][3]);
  LOG_VALUE(D0[i][3][0]);
  LOG_VALUE(D0[i][3][2]);
  LOG_VALUE(dy0);
  LOG_VALUE(Ddy);
  if( !(Ddy!=0) ) return false;

  MpInterval dy = -dy0/Ddy;
  LOG_VALUE(dy);
  LOG_VALUE(r[4]);
  LOG_VALUE(subset(dy,r[4]));

  if(!subset(dy,r[4])) return false;
  // if success, we can replace r[4] by dy as it contains all possible zeroes od P_{dx}
  if(i==this->n-1)
    r[4] = dy*MpInterval(-1,1);
  return true;
}

// #########################################################################
/**
 * This function computes value of reduced bifurcation function g(x,z).
 * Zeroes of g correspond to branch of period-n points.
*/
MpInterval BifurcationTask::computeG(int i){
  intersection(D0[i][5][4] + HBif[i](5,0,4)*rBif[0] + MpInterval(2.)*HBif[i](5,4,4)*rBif[4],DBif[i][5][4],DBif[i][5][4]);
  intersection(D0[i][3][4] + HBif[i](3,0,4)*rBif[0] + MpInterval(2.)*HBif[i](3,4,4)*rBif[4],DBif[i][3][4],DBif[i][3][4]);

  intersection(D0[i][5][0] + MpInterval(2.)*HBif[i](5,0,0)*rBif[0] + HBif[i](5,0,4)*rBif[4],DBif[i][5][0],DBif[i][5][0]);
  intersection(D0[i][3][0] + MpInterval(2.)*HBif[i](3,0,0)*rBif[0] + HBif[i](3,0,4)*rBif[4],DBif[i][3][0],DBif[i][3][0]);

  LOG_VALUE(DBif[i][3][4]);
  if(!(DBif[i][3][4]!=0)) {this->result = false; return MpInterval(0.); }
  
  MpInterval e = -DBif[i][5][4]/DBif[i][3][4];

  MpInterval result = (D0[i][5][0] + e*D0[i][3][0]) + 
                      MpInterval(2.)*(HBif[i](5,0,0)+e*HBif[i](3,0,0))*rBif[0] + 
                      (HBif[i](5,0,4)+e*HBif[i](3,0,4))*rBif[4];
  intersection(result,DBif[i][5][0]+e*DBif[i][3][0],result);
  LOG_VALUE(D0[i][5][0]);
  LOG_VALUE(D0[i][3][0]);
  LOG_VALUE(DBif[i][5][0]);
  LOG_VALUE(DBif[i][3][0]);
  return result;
}

// #########################################################################
/**
 * This function computes dx partial derivative value of the reduced bifurcation function g(x,z).
*/

MpInterval BifurcationTask::computeDxG(int k){
  for(int i=3; i<6;i+=2)
    for(int j=0; j<6;j+=2){
      MpInterval s = D0[k][i][j];
      for(int c=0; c<6;c+=2)
        s += (j==c? 2. : 1.) * H[k](i,j,c)*r[c];
      intersection(D[k][i][j],s,D[k][i][j]);
    }
  LOG_VALUE(D[k][3][4]);
  if(!(D[k][3][4]!=0)) {this->result = false; return MpInterval(0.); }

  solveImplicitEquation(D[k],H[k],DY[k],D2Y[k]);
  MpIHessian _H = H[k]*DY[k] + D[k]*D2Y[k];
  LOG_VALUE(_H(5,0,0));
  LOG_VALUE(_H(5,0,2));
  return _H(5,0,2) + _H(5,0,0)*Dfp[0];
}

// #########################################################################
/**
 * This function verfies, if
 * - the existencecomputes of Lyapunov-Schmidt reduction for all periods less or equal to n
 * - the reduced bifurcation function g_i has no zeroes for i<n
 * - the reduced bifurcation function g_n has branch of zeroes
 * - HC4 condition is satisfied
 * If success, it updates bound on bifurcation point this->r.
*/
bool BifurcationTask::checkBifurcationBranch(){
  MpInterval tmp;
  
  for(int i=0;i<this->n;++i)
    CHECK_AND_LOG("Validate the existence of Lyapunov-Schmidt reduction in the considered domain:\n", computeLyapunovSchmidtReduction(i));
  CHECK_AND_LOG("Intersection of two enclosures of dy(x,z) along fixed point curve:\n",intersection(r[4],fp[4],fp[4]));
  if(!this->result) false;
  rBif=r; rBif[2]=0.;
  DBif = D;
  HBif = H;

//  We check, that g_i for i<n is nonzero. This proves that bifurcating curve has principal period n
  for(int i=0;i<n-1;++i){
    MpInterval g = computeG(i);
    MpInterval Dg = computeDxG(i);
    CHECK_AND_LOG("",Dg!=0.);
    if(!this->result) false;
    // interval Newton operator should be disjoint with the domain
    CHECK_AND_LOG("Check principal period of new branch:\n",!intersection(-g/Dg,r[2],tmp));
    if(!this->result) false;
  }
  MpInterval g = computeG(this->n-1);
  MpInterval Dg = computeDxG(this->n-1);
  if(!this->result) false;

// Check the existence of bifurcation branch for i=n
  LOG_VALUE(g);
  LOG_VALUE(Dg);
  LOG_VALUE(r[2]);
// HC4 condition from the article is eqivalent to Dg!=0
  CHECK_AND_LOG("",Dg!=0.);
  if(!this->result) false;
  LOG_VALUE(-g/Dg);    
  CHECK_AND_LOG("Check existence of period-n branch:\n",subset(-g/Dg,r[2]));
  if(!this->result) false;
  r[2] = -g/Dg;
  return true;
}

// #########################################################################
/**
 * This function takes several intersections
 * to improve bounds on derivatives at bifurcation point.
*/
void BifurcationTask::recomputeC2C3AtBifPoint(){
  MpInterval tmp;

  if(this->n==3){
    this->evalC2(this->cr3bpBifBranchSolver,r,D,H);
    MpIHessian H1 = (H[0]*invM)*D[1] + (D[0]*invM)*H[1];
    MpIMatrix D1 = (D[0]*invM)*D[1];
    LOG_VALUE(D1);
    LOG_VALUE(D[2]);
    CHECK_AND_LOG("Check intersection of two enclosures of hessian at bif point:",intersection(H1,H[2],H[2])); 
    CHECK_AND_LOG("Check intersection of two enclosures of derivative at bif point:",intersection(D1,D[2],D[2])); 
  } else {
    thread c3(&BifurcationTask::evalC3,this),
           c2(&BifurcationTask::evalC2,this,std::ref(cr3bpBifBranchSolver),r,std::ref(D),std::ref(H));
    c2.join(); c3.join();
    if(this->n==4){
      MpIHessian H1 = H[1]*(invM*D[1]) + (D[1]*invM)*H[1];
      MpIMatrix D1 = D[1]*invM*D[1];
      CHECK_AND_LOG("Check intersection of two enclosures of hessian at bif point:",intersection(H1,H[3],H[3])); 
      CHECK_AND_LOG("Check intersection of two enclosures of derivative at bif point:",intersection(D1,D[3],D[3])); 
    }

    if(checkConvexity){ // computation of third order derivatives if the corresponding flag in the config file is set to 1
      J = normalForms::linearSubstitutionPowerSeries(MpIMatrix::Identity(6),J,M);
      // possible improve bound by taking intersection with bound from C2 computation
      for(int i=0;i<6;++i){
        for(int j=0;j<6;++j){
          if(!intersection(J(i,j), D[this->n-1][i][j], tmp)){
            CHECK_AND_LOG("Intersection of bound from C1 and C3 computaion:", intersection(J(i,j), D[this->n-1][i][j], tmp) );
          } else {
            J(i,j) = tmp;
          }
          for(int c=j;c<6;++c){
            if(!intersection(J(i,j,c), H[this->n-1](i,j,c), tmp)){
              CHECK_AND_LOG("Intersection of bound from C2 and C3 computaion:", intersection(J(i,j,c), H[this->n-1](i,j,c), tmp) );
            } else {
              J(i,j,c) = tmp;
            }
          }
        }
      }
    }
  }
}

// #########################################################################

void BifurcationTask::run()
{
  MpFloat::setDefaultPrecision(mantissaBits);
  MpIMatrix DJ(1,6), D1(6,6);
  MpIHessian D2J(1,6), H1(6,6);

  // start clock
  this->t = tic("",this->out);  
  this->init();
  out << "\nProblem: " << msg << " at " << v << endl;

// Step 1. Asynchronous computation of independent tasks.
// 1a. Validate curve of fixed points. If scuccess, fp contains bound on the fixed point curve and Dfp is bound on derivative wrt parameter z.
// 1b. Compute value and derivative of n-th iteration of Poincare map at approximate bifurcation point.
// 1c. Compute derivative and hessian of n-th iteration of Poincare map on the box centred at approximate bifurcation point. 
  thread t1(&BifurcationTask::proveBranchOfFixedPoints,this),
         t2(&BifurcationTask::computeValueAtCentre,this),
         t3(&BifurcationTask::evalC2,this,std::ref(cr3bpBifBranchSolver),r,std::ref(D),std::ref(H));
  t1.join(); t2.join();
  if(!this->result) { this->writeReport(); t3.detach(); t3.~thread(); return; }
  t3.join();
  if(!this->result) { this->writeReport(); return; }
  
// For further computation we restrict the domain in bifurcation direction to the bound on the curve of fixed points
// we multipy by (-1,1) to have our approximate point in the interior of the bound
  fp *= MpInterval(-1,1);
  r[0] = fp[0];

// Step 2. Lyapunov-Schmidt reduction and verification that g_n has branch of zeroes
  if( !this->checkBifurcationBranch() ) { this->writeReport(); return; }

  // recompute C2 and C3 on enclosure of intersection point
  // and improve bounds on higher derivatives of f^n at bifurcation point by taking interesections 
  this->recomputeC2C3AtBifPoint();
  if(!this->result) { this->writeReport(); return; }

  rBif=r; rBif[2]=0.;
  DBif = D;
  HBif = H;

// Finally, we have to recognize type of bifurcation 
// if n is odd then touch-and-go, otherwise period-tupling
  MpIVector bifPointBound = x+M*r;
  MpInterval bifEnergyBound = cr3bpCentreSolver.Jacobi(bifPointBound,DJ,D2J)[0];

  // bound on tangent to fixed point curve at bif point 
  MpIVector T1({Dfp[0],MpInterval(0.),MpInterval(1.),MpInterval(0.),Dfp[1],MpInterval(0.)});
  LOG_VALUE(T1);
  MpIVector DH = (DJ*M).row(0);
  LOG_VALUE(DH[2]);
  CHECK_AND_LOG("Is Jacobi integral monotone along space-parameter of fixed point curve:\n",!isSingular(DH[2]));
  LOG_VALUE(DH*T1);
  CHECK_AND_LOG("Is Jacobi integral monotone along fixed point curve:\n",!isSingular(DH*T1));
  if(!this->result) { this->writeReport(); return; }

  // recompute Dg at bifurcation point
  MpInterval Dg = computeDxG(this->n-1);
  H1 = H[n-1]*DY[n-1] + D[n-1]*D2Y[n-1];
  D1 = D[n-1]*DY[n-1];
  // compute bound on tangent to bifurcation curve at bif point 
  MpInterval dz = -H1(5,0,0)/Dg;
  MpInterval dy = DY[n-1][4][0] + D1[4][2]*dz;
  MpIVector T2({MpInterval(1.),MpInterval(0.),dz,MpInterval(0.),dy,MpInterval(0.)});
  LOG_VALUE(T2);
  LOG_VALUE(DH*T2);

  if(this->n%2){ // touch-and-go case
    CHECK_AND_LOG("Is Jacobi integral monotone along bifurcation curve:\n",!isSingular(DH*T2));
    if(!this->result) { this->writeReport(); return; }
  }
  else          
  // period-tupling case
  {
    CHECK_AND_LOG("Does bound on derivative Jacobi(bifCurve(x)) contain zero:\n",isSingular(DH*T2));
    if(!this->result) { this->writeReport(); return; }

    // second order derivatives of Jacobi integral if the corresponding flag in config file is set to 1
    if(checkConvexity){
      T2 = M*T2;
      MpIMatrix D2H(6,6);
      for(int i=0;i<6;++i)
        for(int j=0;j<6;++j)
          D2H[i][j] = D2J(0,i,j) * (i==j? 2.: 1.);

      MpIJet DY(6,6,3), G(6,6,3);
      solveImplicitEquation(J,DY);
      substitutionPowerSeries(J,DY,G,false);
      MpInterval d2z = -(G(5,0,0,0)*(1.+Dfp[0])/3. + G(5,0,0,2))/Dg;
      MpInterval d2y = 2.*DY(4,0,0) + 2.*DY(4,0,2)*Dfp[0] + 2.*DY(4,2,2)*sqr(dz) + DY(4,2)*d2z;
      MpIVector T3({MpInterval(0.),MpInterval(0.),d2z,MpInterval(0.),d2y,MpInterval(0.)});
      LOG_VALUE(T3);
      LOG_VALUE(DH*T3 + T2*(D2H*T2));
      CHECK_AND_LOG("Is second derivative Jacobi(bifCurve(x)) nonzero:\n",!isSingular(DH*T3 + T2*(D2H*T2)));
      if(!this->result) { this->writeReport(); return; }
    } else{
      out << "Verification of convexity for period-tupling bifurcation is skipped, as indicated by the flag in the config file.\n";
      this->result = false;
    }
  }

  out << "Bound on Jacobi constant: " << bifEnergyBound << endl;
  out << msg << " at " << v << " completed.\nFinal result: " << this-> result << "\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n" << endl;
  this->writeReport();
}

// #########################################################################

int main(int argc, char* argv[])
{   
  MpFloat::setDefaultPrecision(mantissaBits);
  cout.precision(40);
  cout << boolalpha;
  try
  {
    if(argc<2){
      cout << "Usage: ./checkHaloBifurcations pathToDataFile\n";
      cout << "Example:\n./checkHaloBifurcations ./dat/MoonL1_01_PQ.dat\n\n";
      return 0;
    }
    BifurcationTask t(argv[1]);
    t.run();
  }
  catch(exception& e)
  {
    cout << "\n\nException caught: "<< e.what() << endl;
  }
  return 0;
}
