/**
* This file contains common routines for continuation of 
* branches of Halo orbits bifurcating from L1,L2,L3-Lyapunov families.
*/
#include <iostream>
#include <list>
#include "HaloContinuationTask.h"
#include "tictac.h"
#include "ConfigFileReader.h"
typedef std::vector<BaseTask*> TaskArray;

using namespace capd;
using namespace std;

// #########################################################################

// Global variables - values are read from a config file

char filename[128];
int order, cnSolverOrder;
long double tolerance;
double mu;
interval imu;
long double z0,z1;
long double xSizeL, dySizeL;
long double xSizeR, dySizeR;
long double leftZSize, leftZParamSize;
long double rightZSize, rightZParamSize;
long double leftUpperArcSize, leftUpperArcParamSize;
long double rightUpperArcSize, rightUpperArcParamSize;
long double upperArcSplitPoint;

TaskArray tasksUpperArc, tasksZLeft, tasksZRight;

// #########################################################################
/**
 * This function reads constants (global variables) from a config file.
 */ 
void readConfigFile(ConfigFileReader& configFileReader){
  string problem;
  std::cout << configFileReader.getConfigFileContent();
  
  double deltaMu;
  configFileReader.getValue("mu",mu);
  configFileReader.getValue("deltaMu",deltaMu);
  imu = mu+interval(-1,1)*deltaMu;
  CR3BPSolverPool::setMu(mu,imu);

  // read solvers settings
  configFileReader.getValue("order",order);
  configFileReader.getValue("cnSolverOrder",cnSolverOrder);
  CR3BPSolverPool::setOrder(order,cnSolverOrder,cnSolverOrder);
  configFileReader.getValue("tolerance",tolerance);
  CR3BPSolverPool::setTolerance(tolerance);

  configFileReader.getValue("z0",z0);
  configFileReader.getValue("z1",z1);
  configFileReader.getValue("filename",filename);
  HaloSeedPointsReader::readHaloBranch(filename);

  // data for two symmetry breaking bifurcations
  configFileReader.getValue("xSizeL",xSizeL);
  configFileReader.getValue("dySizeL",dySizeL);
  configFileReader.getValue("xSizeR",xSizeR);
  configFileReader.getValue("dySizeR",dySizeR);

  // data for leftZ, rightZ and upperX branches
  configFileReader.getValue("leftZSize",leftZSize);
  configFileReader.getValue("leftZParamSize",leftZParamSize);
  configFileReader.getValue("rightZSize",rightZSize);
  configFileReader.getValue("rightZParamSize",rightZParamSize);
  configFileReader.getValue("leftUpperArcSize",leftUpperArcSize);
  configFileReader.getValue("leftUpperArcParamSize",leftUpperArcParamSize);
  configFileReader.getValue("rightUpperArcSize",rightUpperArcSize);
  configFileReader.getValue("rightUpperArcParamSize",rightUpperArcParamSize);
  configFileReader.getValue("upperArcSplitPoint",upperArcSplitPoint);

  // parameters of the continuation algorithm
  configFileReader.getValue("minParameterStep",HaloContinuationTask::minParameterStep);
  configFileReader.getValue("successSetSizeFactor",HaloContinuationTask::successSetSizeFactor);
  configFileReader.getValue("successParamSizeFactor",HaloContinuationTask::successParamSizeFactor);
  configFileReader.getValue("failureParamSizeFactor",HaloContinuationTask::failureParamSizeFactor);
  configFileReader.getValue("isInitialBoxSizeMaximal",HaloContinuationTask::isInitialBoxSizeMaximal);
  configFileReader.getValue("isAdaptiveSectionChangeAllowed",HaloContinuationTask::isAdaptiveSectionChangeAllowed);
  configFileReader.getValue("testComputation",HaloContinuationTask::testComputation);
}

// #########################################################################
/**
 * This algorithm creates tasks for validation of the existence of Halo branch within parameter range [L,R].
 * The interval [L,R] is split into many parts, each part obtain own seed point for Halo orbit and then it performs continuation.
 * @param [L,R] - interval of parameters
 * @param c - index of variable which is the parameter - x, z or dy.
 * @param i,j - indexes of two variables which are not parameters
 * @param si,sj - initial sizes of box in (i,j)-variables for Interval Krawczyk Operator
 * @param step - initial step in c-variable for continuation
 * @param delta - size of subinterval of [L,R] to be processed by a single task
 * @param predicate - used to select points from file with precomputed seed points 
*/ 
void createTasks(
    std::vector<BaseTask*>& tasks,
    double L, double R,
    int i, int j, int c,
    double si, double sj, double step, double delta, 
    Predicate predicate, bool progresive=false
){
  double SI = si, SJ=sj, D=R-L;
  do{
    LDVector seed = HaloSeedPointsReader::getHaloSeedPoint(predicate,L,i,j,c);
    double R0 = capd::min(R,L+delta);
    if(progresive){
      double t = (R-L)/D;
      SJ = SI = sj + t*(si-sj);
    }
    tasks.push_back(new HaloContinuationTask(interval(L,R0),step,tasks.size()+1,i,j,c,SI,SJ,seed));    
    L = R0;
  }while(L<R);
}

// #########################################################################

bool proveSymBreakingBifurcation(interval x, interval dy, double sizeX, double sizeDY, long double z0){
  ICR3BP  icr3bp(20,1e-18);
  IPCR3BP ipcr3bp(20,1e-18);
  interval I = interval(-1,1);
  IVector lbox({I*sizeX,0.,0.,I*sizeDY});
  IVector hbox({I*sizeX,0.,interval(0,1)*z0,0.,I*sizeDY,0.});
  IVector halo({x,0.,0.,0.,dy,0.});
  IVector planar({x,0.,0.,dy});
  // note, we do not prove convexity, as it was already validated in more general computation
  // regarding symmetry breaking bifurcations for wide range of mu.
  return proveLyapunovOrbit(ipcr3bp, planar,lbox,imu) and proveBifHaloOrbit(icr3bp, halo,hbox,imu);
}
