#include "checkHaloContinuation.h"

// #########################################################################

void createTasksForZArcs(){
  Predicate leftBranch =  [](long double p, long double x, long double z, long double dy){ return x>0 or z>p; },
            rightBranch = [](long double p, long double x, long double z, long double dy){ return z<p; };

  double z = 1./512;
  createTasks(tasksZLeft, z0,z, 0,4,2, leftZSize,leftZSize,leftZParamSize, 1./4096, leftBranch);  
  createTasks(tasksZLeft, z,z1, 0,4,2, leftZSize,leftZSize,leftZParamSize, 1./1024, leftBranch);  
  createTasks(tasksZRight,z0,z1, 0,4,2, rightZSize,rightZSize,rightZParamSize, 1./1536, rightBranch);
}

// #########################################################################
  
void createTasksForUpperArc(long double x0, long double x1){
  Predicate upperBranch = [](long double p, long double x, long double z, long double dy){ return x>p or z<0.3; };

  createTasks(tasksUpperArc, x0,upperArcSplitPoint, 2,4,0, leftUpperArcSize,leftUpperArcSize,leftUpperArcParamSize, 1./1024, upperBranch); 
  createTasks(tasksUpperArc, upperArcSplitPoint,x1, 2,4,0, rightUpperArcSize,rightUpperArcSize,rightUpperArcParamSize, 1./1536, upperBranch); 
}

// #########################################################################
/**
 * This function realizes a computer-assisted proof
 * that the family of Halo orbits arising near L1 forms a circle.
 *
 * We have to solve the following issues:
 * 1. There are two symmetry breaking bifurcations at the points where the circle corsses x-axis
 * 2. The curve cannot be parameterized by just one of the variables x,z,dy
 *
 * In order to get a continuous half-circle we proceed as follows:
 * 1. We prove the existence of two symmetry breaking bifurcations. This allows to prove the existence of Halo orbits for z\in [0,z0]
 * 2. We continue two branches starting from symmetry breaking bifurcations until some fixed z=z1.
 *    The two branches are located in x<0 and x>0 subspace, respectively.
 * 3. We join these two parts by a branch parameterized by x-variable.
 * 4. Eventually we check that all pieces glue into a smooth curve.
 */ 
void proveL1HaloBranch(){
  DCR3BP& dcr3bp = CR3BPSolverPool::cr3bp->getSolver(0);
  LDVector rBif =  findBifHaloOrbit(dcr3bp,*HaloSeedPointsReader::getHaloBranch().begin());
  LDVector lBif =  findBifHaloOrbit(dcr3bp,*HaloSeedPointsReader::getHaloBranch().rbegin());
  cout << "Left symmetry breaking bifurcation:\n" << lBif << endl;
  cout << "Right symmetry breaking:\n" << rBif << endl;
  // Check the existence of two symmetry breaking bifurcations (in x<0 and x>0 subspaces, respectively).
  // From this we get the existence of Halo orbits for z\in[0,z0].
  // Remark: we do not check nondegeneracy condition (convexity of bifurcating curve) because it was already checked 
  // in more general computation regarding symmetry breaking bifurcations for wide range of mu.
  CHECK_AND_PRINT("Check two symmetry breaking bifurcations:\n",
    proveSymBreakingBifurcation(lBif[0], lBif[4], xSizeL, dySizeL, z0) and
    proveSymBreakingBifurcation(rBif[0], rBif[4], xSizeR, dySizeR, z0)
  );
  createTasksForZArcs();
// Step 2a.
  // Compute arc of Halo orbits parameterized by z\in[z0,z1] in x<0 subspace.
  cout << "\n\nCompute left z-arc of L1 Halo orbits.\n"; 
  BaseTask::runTasks(tasksZLeft,BaseTask::FORWARD);
  
// Step 2b.
  // Compute arc of Halo orbits parameterized by z\in[z0,z1] in x>0 subspace.
  cout << "\n\nCompute right z-arc of L1 Halo orbits.\n"; 
  BaseTask::runTasks(tasksZRight,BaseTask::FORWARD);

// Step 3.
  // Compute upper arc parameterized by x-coordinate.
  // We start from x that belongs to the last set in the left z-arc
  // We stop at x that belongs to the last set in the right z-arc
  cout << "\n\nCompute upper x-arc of L1 Halo orbits.\n"; 
  IVector hLeft  = static_cast<HaloContinuationTask*>(tasksZLeft.back())->boundLast.get_x();
  IVector hRight  = static_cast<HaloContinuationTask*>(tasksZRight.back())->boundLast.get_x();
  cout << "x0=" << hLeft[0].leftBound() << endl;
  cout << "x1=" << hRight[0].rightBound() << endl;
  createTasksForUpperArc(hLeft[0].leftBound(),hRight[0].rightBound());
  BaseTask::runTasks(tasksUpperArc,BaseTask::FORWARD);

// Step 4.
  // We check that all the pieces glue into a continuous curve
  // First we check that all pieces z-left, z-right and x-branch are continuous
  if(!HaloContinuationTask::testComputation){
    CHECK_AND_PRINT("Check if left z-arc is continuous:\n", checkNonemptyIntersection(tasksZLeft,0,4,2));
    CHECK_AND_PRINT("Check if right z-arc is continuous:\n",checkNonemptyIntersection(tasksZRight,0,4,2));
    CHECK_AND_PRINT("Check if upper x-arc is continuous:\n",checkNonemptyIntersection(tasksUpperArc,2,4,0));
  }
// Step 4a.
  // here we check if arc from symmetry breaking bifurcation meets arc from continuation for z=z0.
  // Since both methods guarantee uniquenes of x(z), dy(z) it is enough to check that the bounds overlap.
  IVector boundLeft = static_cast<HaloContinuationTask*>(tasksZLeft.front())->boundFirst;
  CHECK_AND_PRINT("Check if left z-branch joins symmetry breaking bifurcation:\n",
    subset(boundLeft[0],lBif[0]  + interval(-1,1)*xSizeL ) and
    subset(boundLeft[4],lBif[4] + interval(-1,1)*dySizeL)
  ); 

  IVector boundRight = static_cast<HaloContinuationTask*>(tasksZRight.front())->boundFirst;
  CHECK_AND_PRINT("Check if right z-arc joins symmetry breaking bifurcation:\n",
    subset(boundRight[0],rBif[0] + interval(-1,1)*xSizeR ) and
    subset(boundRight[4],rBif[4] + interval(-1,1)*dySizeR)
  );

// Step 4c. Intersection of x-arc with z-left and z-right arcs. 
  // First we compute extra Halo orbit at the end of x-arc
  IVector boxRight(6);
  ICR3BP&  icr3bp  = CR3BPSolverPool::icr3bp->getSolver(0);
  icr3bp.setSection(1);
  CHECK_AND_PRINT("",proveHaloOrbitAtASinglePoint(icr3bp,hRight,boxRight,2,4,0,3,5,1e-6));
  CHECK_AND_PRINT("",proveHaloOrbit(icr3bp,hRight,2,4,1e-10,1e-10, boxRight,3,5));

  // Then we check that there is a Halo orbit that belongs to the intersection of
  // z-left\cap x-upper and z-right\cap x-upper.
  // This proves that these three curves form one smooth curve. 
  HaloContinuationTask* lX = static_cast<HaloContinuationTask*>(tasksUpperArc.front());
  HaloContinuationTask* rX = static_cast<HaloContinuationTask*>(tasksUpperArc.back());
  HaloContinuationTask* lZ = static_cast<HaloContinuationTask*>(tasksZLeft.back());
  HaloContinuationTask* rZ = static_cast<HaloContinuationTask*>(tasksZRight.back());

  CHECK_AND_PRINT("Check if left z-arc glues with x-arc:\n",
    checkInclusion(IVector(lX->boundFirst),lZ->boundLast)
  );
  CHECK_AND_PRINT("Check if right z-arc glues with x-arc:\n",
    checkNonemptyIntersection(AffineSet(hRight,boxRight),rX->boundLast,2,4,0,hRight[0].rightBound()) and
    checkInclusion(hRight+boxRight,rZ->boundLast)
  );

  BaseTask::releaseTasksMemory(tasksUpperArc);
  BaseTask::releaseTasksMemory(tasksZLeft);
  BaseTask::releaseTasksMemory(tasksZRight);
}

// #########################################################################

int main(int argc, char* argv[]){

  if(argc>1){
    cout.precision(17);
    CR3BPSolverPool::initSolvers();

    ConfigFileReader reader(argv[1]);
    readConfigFile(reader);
    
    auto t = tic("");
    proveL1HaloBranch();
    tac(t);

    CR3BPSolverPool::releaseSolvers();
  } else {
    cout << "Usage:\n./checkL1HaloContinuation pathToConfigFile\n\nExamples:\n";
    cout << "./checkL1HaloContinuation ./conf/SunJupiterL1Branch.conf\n";
    cout << "./checkL1HaloContinuation ./conf/EarthMoonL1Branch.conf\n\n";
  }

  return 0;
}


