#include "checkHaloContinuation.h"

// #########################################################################

void createTasksForZArc(){
  Predicate rightBranch = [](long double p, long double x, long double z, long double dy){ return x>1 and z<p; };
  double z = 1./8192;
  createTasks(tasksZRight,z0,z, 0,4,2, rightZSize,rightZSize,rightZParamSize, 1./524288, rightBranch);
  createTasks(tasksZRight,z,z1, 0,4,2, rightZSize,rightZSize,rightZParamSize, 1./8192, rightBranch);
  for(auto& _t : tasksZRight){
    HaloContinuationTask* t= static_cast<HaloContinuationTask*>(_t); 
    t->implicitFunctionDerivative = new IVector(2);
  }
}

// #########################################################################
  
void createTasksForUpperArc(long double dy0, long double dy1){
  Predicate upperBranch = [](long double p, long double x, long double z, long double dy){ return dy<p; };
  createTasks(tasksUpperArc, dy0,dy1, 0,2,4, leftUpperArcSize,rightUpperArcSize,leftUpperArcParamSize, 1./128, upperBranch,true); 
}

// #########################################################################
/**
 * This function realizes a computer-assisted proof
 * that the family of Halo orbits arising near L2 is a parabola.
 *
 * We have to solve the following issues:
 * 1. There is one symmetry breaking bifurcation
 * 2. The curve cannot be parameterized by dy. Initialy we parametrize it by z and then switch to dy
 *
 * In order to get one branch we proceed as follows:
 * 1. We prove the existence of a symmetry breaking bifurcation. This allows to prove the existence of Halo orbits for z\in [0,z0]
 * 2. We continue this branch starting from symmetry breaking bifurcation until some fixed z=z1.
 * 3. We join this branch with another segment parametrized by dy.
 * 4. Eventually we check that all pieces glue into a smooth curve.
 */ 
void proveHaloBranch(){
  DCR3BP& dcr3bp = CR3BPSolverPool::cr3bp->getSolver(0);
  LDVector rBif =  findBifHaloOrbit(dcr3bp,*HaloSeedPointsReader::getHaloBranch().begin());

// Step 1.
  // Check the existence of symmetry breaking bifurcation.
  // From this we get the existence of Halo orbits for z\in[0,z0].
  // Remark: we do not check nondegeneracy condition (convexity of bifurcating curve) because it was already checked 
  // in more general computation regarding symmetry breaking bifurcations for wide range of mu.
  CHECK_AND_PRINT("Check symmetry breaking bifurcation:\n",
    proveSymBreakingBifurcation(rBif[0], rBif[4], xSizeR, dySizeR, z0)
  );

// Step 2.
  // Compute arc of Halo orbits parameterized by z\in[z0,z1].
  createTasksForZArc();
  cout << "\n\nCompute z-arc of halo orbits.\n"; 
  BaseTask::runTasks(tasksZRight,BaseTask::FORWARD);

// Step 3.
  // Compute arc parameterized by dy-coordinate.
  // We start from dy that belongs to the last set in the right z-arc
  // We stop at dy that belongs to the last set in the right z-branch
  cout << "\n\nCompute dy-arc of halo orbits.\n"; 
  IVector hRight  = static_cast<HaloContinuationTask*>(tasksZRight.back())->boundLast.get_x();
  cout << "dy0=" << hRight[4].rightBound() << endl;
  createTasksForUpperArc(hRight[4].leftBound(),upperArcSplitPoint);
  BaseTask::runTasks(tasksUpperArc,BaseTask::FORWARD);

// Step 4.
  // We check that all the pieces join into a continuous curve
  // First we check that all pieces z-branch and dy-branch are continuous
  if(!HaloContinuationTask::testComputation){
    CHECK_AND_PRINT("Check if z-arc is continuous:\n",checkNonemptyIntersection(tasksZRight,0,4,2));
    CHECK_AND_PRINT("Check if dy-arc is continuous:\n",checkNonemptyIntersection(tasksUpperArc,0,2,4));
  }

// Step 4a.
  // here we check if branch from symmetry breaking bifurcation meets branch from continuation for z=z0.
  // Since both methods guarantee uniquenes of x(z), dy(z) it is enough to check that the bounds overlap.
  IVector boundRight = static_cast<HaloContinuationTask*>(tasksZRight.front())->boundFirst;
  CHECK_AND_PRINT("Check if z-branch joins symmetry breaking bifurcation:\n",
    subset(boundRight[0],rBif[0] + interval(-1,1)*xSizeR ) and
    subset(boundRight[4],rBif[4] + interval(-1,1)*dySizeR)
  );

// Step 4c. Intersection of dy-branch with z-branch. 
  // First we compute extra Halo orbit at the end of dy-branch
  IVector boxRight(6);
  ICR3BP&  icr3bp  = CR3BPSolverPool::icr3bp->getSolver(0);
  icr3bp.setSection(1);
  CHECK_AND_PRINT("",proveHaloOrbit(icr3bp,hRight,2,4,1e-10,1e-10, boxRight,3,5));

  // Then we check that there is a halo orbit that belongs to the intersection of
  // z-branch\cap dy-branch. From the uniqueness this proves,
  // that these two curves form one smooth curve. 
  HaloContinuationTask* dy = static_cast<HaloContinuationTask*>(tasksUpperArc.front());
  HaloContinuationTask* z = static_cast<HaloContinuationTask*>(tasksZRight.back());

  cout << "first=" << IVector(dy->boundFirst) << endl;
  cout << "last=" << IVector(z->boundLast) << endl;
  CHECK_AND_PRINT("Check if z-arc joins dy-arc:\n",
    checkInclusion(IVector(dy->boundFirst),z->boundLast)
  );

  BaseTask::releaseTasksMemory(tasksUpperArc);
  BaseTask::releaseTasksMemory(tasksZRight);
}

// #########################################################################

int main(int argc, char* argv[]){

  if(argc>1){
    cout.precision(17);
    CR3BPSolverPool::initSolvers();

    ConfigFileReader reader(argv[1]);
    readConfigFile(reader);

    auto t = tic("");
    proveHaloBranch();
    tac(t);

    CR3BPSolverPool::releaseSolvers();
  } else {
    cout << "Usage:\n./checkL2HaloContinuation pathToConfigFile\n\nExamples:\n";
    cout << "./checkL2HaloContinuation ./conf/SunJupiterL2Branch.conf\n";
    cout << "./checkL2HaloContinuation ./conf/EarthMoonL2Branch.conf\n\n";
  }

  return 0;
}


