#include "CR3BPSolver.h"
#include "capd/dynset/C0HOSet.hpp"
#include "capd/dynset/C1HOSet.hpp"
#include "capd/dynset/C0DoubletonSet.hpp"
#include "capd/dynset/C1DoubletonSet.hpp"
#include "capd/dynset/C2DoubletonSet.hpp"
#include "capd/dynset/CnRect2Set.hpp"
#include "capd/dynsys/CnOdeSolver.hpp"
using namespace capd;

template class capd::dynset::C1DoubletonSet<capd::IMatrix,PartialQR>;
template class capd::dynset::C1HOSet<PC1Rect2Set>;
template class capd::dynset::C2DoubletonSet<capd::IMatrix,PartialQR>;
template class capd::dynset::CnRect2Set<capd::IMatrix,PartialQR>;
template class capd::dynsys::CnOdeSolver<capd::IMap,capd::dynsys::ILastTermsStepControl>;

using capd::autodiff::Node;
/**
  This routine implements vector field of the PCR3BP.
  The routine is written in the format required by the constructor
  of the class DMap and IMap from the CAPD library.
*/
void pcr3bpVectorField(Node t, Node in[], int dimIn, Node out[], int dimOut, Node params[], int noParams){
  Node mj = 1-params[0]; // relative mass of the second body
  Node xMu = in[0] + params[0];
  Node xMj = in[0] - mj;
  Node xMuSquare = xMu^2; // square
  Node xMjSquare = xMj^2;
  Node ySquare = in[1]^2;

  Node factor1 = mj*((xMuSquare+ySquare)^-1.5);
  Node factor2 = params[0]*((xMjSquare+ySquare)^-1.5);

  out[0] = in[2];
  out[1] = in[3];
  out[2] = in[0] - xMu*factor1 - xMj*factor2 + 2*in[3];
  out[3] = in[1]*(1 - factor1 - factor2) - 2*in[2];
}
  
/**
  This routine implements vector field of the PCR3BP.
  The routine is written in the format required by the constructor
  of the class DMap and IMap from the CAPD library.
*/
void extendedPcr3bpVectorField(Node t, Node in[], int dimIn, Node out[], int dimOut, Node params[], int noParams){
  Node mj = 1-in[4]; // relative mass of the second body
  Node xMu = in[0] + in[4];
  Node xMj = in[0] - mj;
  Node xMuSquare = xMu^2; // square
  Node xMjSquare = xMj^2;
  Node ySquare = in[1]^2;

  Node factor1 = mj*((xMuSquare+ySquare)^-1.5);
  Node factor2 = in[4]*((xMjSquare+ySquare)^-1.5);

  out[0] = in[2];
  out[1] = in[3];
  out[2] = in[0] - xMu*factor1 - xMj*factor2 + 2*in[3];
  out[3] = in[1]*(1 - factor1 - factor2) - 2*in[2];
  out[4] = Node(0.0);
}

  // #########################################################################

/**
  This routine implements vector field of the CR3BP.
  The routine is written in the format required by the constructor
  of the class DMap and IMap from the CAPD library.
*/
void cr3bpVectorField(Node t, Node in[], int dimIn, Node out[], int dimOut, Node params[], int noParams){
  Node xMu = in[0] + params[0]; // params[0] - relative mass of the first body
  Node xMj = in[0] + params[1]; // params[1] = params[0] - 1
  Node xMuSquare = xMu^2;       // square
  Node xMjSquare = xMj^2;
  Node ySquare = in[1]^2;
  Node zSquare = in[2]^2;
  Node yzSquare = ySquare + zSquare;
  
  Node factor1 = params[1]*((xMuSquare+yzSquare)^-1.5);
  Node factor2 = params[0]*((xMjSquare+yzSquare)^-1.5);
  Node factor = factor1 - factor2;

  out[0] = in[3];
  out[1] = in[4];
  out[2] = in[5];

  out[3] = in[0] + xMu*factor1 - xMj*factor2 + 2*in[4];
  out[4] = in[1]*(1+factor) - 2*in[3];
  out[5] = in[2]*factor;
}

// #########################################################################

/**
  This routine implements extended vector field of the CR3BP with mass (mu) as a space variable.
  Such approach allows to compute derivateves with respect to mu.
*/
void extendedCr3bpVectorField(Node t, Node in[], int dimIn, Node out[], int dimOut, Node params[], int noParams){
  Node mj = in[6] - 1;
  Node& mu = in[6];
  
  Node xMu = in[0] + mu; // params[0] - relative mass of the first body
  Node xMj = in[0] + mj; // params[1] = params[0] - 1
  Node xMuSquare = xMu^2;       // square
  Node xMjSquare = xMj^2;
  Node ySquare = in[1]^2;
  Node zSquare = in[2]^2;
  Node yzSquare = ySquare + zSquare;
  
  Node factor1 = mj*((xMuSquare+yzSquare)^-1.5);
  Node factor2 = mu*((xMjSquare+yzSquare)^-1.5);
  Node factor = factor1 - factor2;

  out[0] = in[3];
  out[1] = in[4];
  out[2] = in[5];

  out[3] = in[0] + xMu*factor1 - xMj*factor2 + 2*in[4];
  out[4] = in[1]*(1+factor) - 2*in[3];
  out[5] = in[2]*factor;

  out[6] = Node(0.0);
}

// #########################################################################

/**
  This routine implements computation of Jacobi first integral for 3D system.
*/
void JacobiIntegral3D(Node t, Node in[], int dimIn, Node out[], int dimOut, Node params[], int noParams)
{
  Node mu = params[0];
  Node mj = 1-mu;
  Node xMu = in[0] + mu;
  Node xMj = in[0] - mj;
  Node xMuSquare = sqr(xMu);
  Node xMjSquare = sqr(xMj);
  Node ySquare = sqr(in[1]);
  Node zSquare = sqr(in[2]);
  
  Node factor1 = mj/sqrt(xMuSquare+ySquare+zSquare);
  Node factor2 = mu/sqrt(xMjSquare+ySquare+zSquare);

  out[0] =  sqr(in[0]) + ySquare - (sqr(in[3]) + sqr(in[4]) + sqr(in[5])) + 2*(factor1 + factor2);
}

// #########################################################################

/**
  This routine implements computation of Jacobi first integral for 2D system.
*/
void JacobiIntegral2D(Node t, Node in[], int dimIn, Node out[], int dimOut, Node params[], int noParams)
{
  Node mu = params[0];
  Node mj = 1-mu;
  Node xMu = in[0] + mu;
  Node xMj = in[0] - mj;
  Node xMuSquare = sqr(xMu);
  Node xMjSquare = sqr(xMj);
  Node ySquare = sqr(in[1]);
  
  Node factor1 = mj/sqrt(xMuSquare+ySquare);
  Node factor2 = mu/sqrt(xMjSquare+ySquare);

  out[0] =  sqr(in[0]) + ySquare - (sqr(in[2]) + sqr(in[3])) + 2*(factor1 + factor2);
}

// #########################################################################

DPCR3BP::DPCR3BP(int order) 
  : vf(pcr3bpVectorField,4,4,1,3),
    exvf(extendedPcr3bpVectorField,5,5,0),
    solver(vf,order,LDOdeSolver::StepControlType(2,1e-12)),
    exsolver(exvf,order,LDOdeSolver::StepControlType(2,1e-12)),
    section(4,1),
    exsection(5,1),
    pm(solver,section),
    expm(exsolver,exsection),
    mu(0.5)
{
  vf.setParameter(0,mu);
}

void DPCR3BP::setMu(long double mu){
  this->mu = mu;
  this->vf.setParameter(0,mu);
}

// #########################################################################

DCR3BP::DCR3BP(int order) 
  : vf(cr3bpVectorField,6,6,2),
    exvf(extendedCr3bpVectorField,7,7,0),
    solver(vf,order,LDOdeSolver::StepControlType(2,1e-12)),
    exsolver(exvf,order,LDOdeSolver::StepControlType(2,1e-12)),
    y(6,1), dx(6,3), dz(6,5), exsection(7,1),
    pm(solver,y),
    expm(exsolver,exsection),
    mu(0.5)
{
  vf.setParameter(0,mu);
  vf.setParameter(1,mu-1);
}

void DCR3BP::setMu(long double mu){
  this->mu = mu;
  this->vf.setParameter(0,mu);
  this->vf.setParameter(1,mu-1);
}

void DCR3BP::setSection(int index){
  switch(index){
    case 1: pm.setSection(y); break;
    case 3: pm.setSection(dx); break;
    case 5: pm.setSection(dz); break;
    default: throw std::runtime_error("Incorrect index in DCR3BP::setSection");  
  }
}
// #########################################################################

MpCR3BP::MpCR3BP(int order, double tol) 
  : vf(cr3bpVectorField,6,6,2),
    solver(vf,order),
    section(6,1),
    pm(solver,section),
    mu(0.5)
{
  vf.setParameter(0,mu);
  vf.setParameter(1,mu-1.);
  solver.setRelativeTolerance(tol);
  solver.setAbsoluteTolerance(tol);
}

void MpCR3BP::setMu(capd::MpFloat mu){
  this->mu = mu;
  this->vf.setParameter(0,mu);
  this->vf.setParameter(1,mu-1.);
}

// #########################################################################

IPCR3BP::IPCR3BP(int order, double tol) 
  : vf(pcr3bpVectorField,4,4,1),
    exvf(extendedPcr3bpVectorField,5,5,0),
    Jacobi(JacobiIntegral2D,4,1,1,2),
    solver(vf,order,capd::IOdeSolver::StepControlType(2,1e-12)),
    exsolver(exvf,order,IOdeSolver::StepControlType(2,1e-12)),
    section(4,1),
    exsection(5,1),
    pm(solver,section),
    expm(exsolver,exsection)
{
  vf.setParameter(0,0.5);
  Jacobi.setParameter(0,0.5);
  solver.setAbsoluteTolerance(tol);
  solver.setRelativeTolerance(tol);
  exsolver.setAbsoluteTolerance(tol);
  exsolver.setRelativeTolerance(tol);
}

void IPCR3BP::setMu(capd::interval mu){
  this->vf.setParameter(0,mu);
  this->Jacobi.setParameter(0,mu);
}

// #########################################################################

ICR3BP::ICR3BP(int order, double tol) 
  : vf(cr3bpVectorField,6,6,2),
    c2vf(cr3bpVectorField,6,6,2,2),
    cnvf(cr3bpVectorField,6,6,2,3),
    exvf(extendedCr3bpVectorField,7,7,0),
    exc2vf(extendedCr3bpVectorField,7,7,0,2),
    excnvf(extendedCr3bpVectorField,7,7,0,3),
    Jacobi(JacobiIntegral3D,6,1,1,2),
    y(6,1), dx(6,3), dz(6,5), exsection(7,1),
    solver(vf,order,IOdeSolver::StepControlType(2,1e-12)),
    exsolver(exvf,order,IOdeSolver::StepControlType(2,1e-12)),
    c2solver(c2vf,15), exc2solver(exc2vf,15),
    cnsolver(cnvf,20), excnsolver(excnvf,20),
    pm(solver,y),
    expm(exsolver,exsection),
    c2pm(c2solver,y),
    exc2pm(exc2solver,exsection),
    cnpm(cnsolver,y),
    excnpm(excnsolver,exsection),
    mu(0.5)
{
  vf.setParameter(0,0.5);
  vf.setParameter(1,0.5);
  c2vf.setParameter(0,0.5);
  c2vf.setParameter(1,0.5);
  cnvf.setParameter(0,0.5);
  cnvf.setParameter(1,0.5);
  Jacobi.setParameter(0,0.5);

  solver.setAbsoluteTolerance(tol);
  solver.setRelativeTolerance(tol);
  c2solver.setAbsoluteTolerance(tol);
  c2solver.setRelativeTolerance(tol);
  c2solver.setStepControl(IC2OdeSolver::StepControlType(2,1e-12));
  cnsolver.setAbsoluteTolerance(1e-18);
  cnsolver.setRelativeTolerance(1e-18);
 // cnsolver.setStepControl(ICnOdeSolver::StepControlType(2,1e-18));

  exsolver.setAbsoluteTolerance(tol);
  exsolver.setRelativeTolerance(tol);
  exc2solver.setAbsoluteTolerance(tol);
  exc2solver.setRelativeTolerance(tol);
  exc2solver.setStepControl(IC2OdeSolver::StepControlType(2,1e-12));
  excnsolver.setAbsoluteTolerance(1e-18);
  excnsolver.setRelativeTolerance(1e-18);
}

void ICR3BP::setMu(capd::interval mu){
  this->mu = mu;
  this->vf.setParameter(0,mu);
  this->vf.setParameter(1,mu-1);
  this->c2vf.setParameter(0,mu);
  this->c2vf.setParameter(1,mu-1);
  this->cnvf.setParameter(0,mu);
  this->cnvf.setParameter(1,mu-1);
  this->Jacobi.setParameter(0,mu);
}
  
void ICR3BP::setSection(int index){
  switch(index){
    case 1: pm.setSection(y);  c2pm.setSection(y);  cnpm.setSection(y);  break;
    case 3: pm.setSection(dx); c2pm.setSection(dx); cnpm.setSection(dx); break;
    case 5: pm.setSection(dz); c2pm.setSection(dz); cnpm.setSection(dz); break;
    default: throw std::runtime_error("Incorrect index in ICR3BP::setSection");  
  }
}

// #########################################################################

MpICR3BP::MpICR3BP(int order, double tol) 
  : vf(cr3bpVectorField,6,6,2), cnvf(cr3bpVectorField,6,6,2,3),
    Jacobi(JacobiIntegral3D,6,1,1,2),
    section(6,1),
    solver(vf,order),
    pm(solver,section),
    cnsolver(cnvf,order),
    cnpm(cnsolver,section),
    mu(0.5)
{
  vf.setParameter(0,mu);
  vf.setParameter(1,mu);
  cnvf.setParameter(0,mu);
  cnvf.setParameter(1,mu);
  Jacobi.setParameter(0,mu);
  solver.setAbsoluteTolerance(tol);
  solver.setRelativeTolerance(tol);
}

void MpICR3BP::setMu(capd::MpInterval mu){
  this->mu = mu;
  this->Jacobi.setParameter(0,mu);
  this->vf.setParameter(0,mu);
  this->vf.setParameter(1,mu-1.);
  this->cnvf.setParameter(0,mu);
  this->cnvf.setParameter(1,mu-1.);
}

void MpICR3BP::setOrderAndToleranve(int order, double tol){
  solver.setOrder(order);
  solver.setAbsoluteTolerance(tol);
  solver.setRelativeTolerance(tol);
  cnsolver.setAbsoluteTolerance(tol);
  cnsolver.setRelativeTolerance(tol);
} 
