#include "HaloContinuationTask.h"
using namespace capd;
using namespace std;

// class static members
double HaloContinuationTask::minParameterStep = 1e-12;
double HaloContinuationTask::successSetSizeFactor = 1.1;
double HaloContinuationTask::successParamSizeFactor = 1.005;
double HaloContinuationTask::failureParamSizeFactor = 0.9;
bool HaloContinuationTask::isInitialBoxSizeMaximal = true;
bool HaloContinuationTask::isAdaptiveSectionChangeAllowed = false;
bool HaloContinuationTask::testComputation = false;

// ######################################################################

HaloContinuationTask::HaloContinuationTask(interval param, double stepParam, int taskNo, int i, int j, int c, double si, double sj, LDVector seed)
  : BaseTask(param,stepParam,taskNo),
    i(i), j(j), c(c),
    sizeI(si), sizeJ(sj),
    SI(-si,si), SJ(-sj,sj),
    seed(seed),
    boundFirst(6), boundLast(6), bound(6)
{}

// ######################################################################

/**
 * Validation of Halo orbit for one parameter value.
 * Then we start continuation from this point.
*/
bool proveHaloOrbitAtASinglePoint(ICR3BP& icr3bp, IVector halo, IVector& box, int i, int j, int c, int pi1, int pi2, double size){
  double si = 1e-11;
  double sj = 1e-11;
  bool result = false;

  // first, we try to validate Halo orbit on a single parameter value
  do{
    try{
      result = proveHaloOrbit(icr3bp,halo,i,j,si,sj,box,pi1,pi2);
    } catch(...) {}
    si*=1.5;
    sj*=1.5;
  }while(si<size and !result);

  return result;
}

// ######################################################################


bool HaloContinuationTask::initComputation(ICR3BP& icr3bp, IVector halo, IVector& box){
  return proveHaloOrbitAtASinglePoint(icr3bp,halo,box,i,j,c,pi1,pi2,sizeI);
}
  
// ######################################################################
/**
 * The function sets initial box size for further continuation.
 * Can be overriden in inheriting class.
*/
bool HaloContinuationTask::acceptEnclosure(const AffineSet& bound, double param){
  if(implicitFunctionDerivative!=nullptr){
    interval dy = bound.getElement_B(4,2) + (*implicitFunctionDerivative)[1];
    if(isSingular(dy)){
      PRINT( (*implicitFunctionDerivative) );
      return false;
    }
  }
  return checkNonemptyIntersection(this->boundLast,bound,this->i,this->j,this->c,param);
}

// ######################################################################

/**
 * The function selects one of the section y=0, dx=0, dz=0
 * for validation of Halo orbits on a range of parameters.
*/
void HaloContinuationTask::selecSection(ICR3BP& icr3bp, DCR3BP& dcr3bp){
  this->pi1=3; this->pi2=5;
  if(isAdaptiveSectionChangeAllowed){
    LDVector f = dcr3bp.vf(dcr3bp.pm(seed));
    if(fabs(f[1])>=fabs(f[3]) and fabs(f[1])>=fabs(f[5])){
      dcr3bp.setSection(1);
      icr3bp.setSection(1);
      this->pi1 = 3; this->pi2 = 5;
    } else if(fabs(f[3])>=fabs(f[1]) and fabs(f[3])>=fabs(f[5])){
      dcr3bp.setSection(3);
      icr3bp.setSection(3);
      this->pi1 = 1; this->pi2 = 5;
    } else {
      dcr3bp.setSection(5);
      icr3bp.setSection(5);
      this->pi1 = 1; this->pi2 = 3;
    }
  }
}

// ######################################################################

/**
 * Validation of Halo orbit on a range of parameters.
 * This virtual method is executed as a task by ThreadPool.
*/
void HaloContinuationTask::run(unsigned id){    
  double first = this->parameterRange.leftBound();
  double last = this->parameterRange.rightBound();
  double current = first;
  this->validatedRange = first;
  
  ICR3BP& icr3bp = CR3BPSolverPool::icr3bp->getSolver(id);
  DCR3BP& dcr3bp = CR3BPSolverPool::cr3bp->getSolver(id);
  this->selecSection(icr3bp,dcr3bp);

  LDVector h = findHaloOrbit(dcr3bp,seed,i,j,pi1,pi2);
  IVector halo = vectalg::convertObject<IVector>(h);
  const interval successBoxScalefactor = HaloContinuationTask::successSetSizeFactor*interval(-1,1);

  IVector box(6);
  HaloContinuationTask* t = nullptr;

  if(initComputation(icr3bp,halo,box)){
    boundFirst = AffineSet(halo,box);
    boundLast = boundFirst;
    this->si = this->SI;
    this->sj = this->SJ;
    
    do{  
      double next = capd::min(current + this->stepParameter,last);
      interval sc = interval(current,next);
      bool oneStepResult = false;
      interval trialSI = si;
      interval trialSJ = sj;
      try{
        h[c] = this->getSeedParameter(current,next);
        h = findHaloOrbit(dcr3bp,h,i,j,pi1,pi2);
        // Call algorithm from halo.cpp
        oneStepResult = proveBranchHaloOrbits(dcr3bp,icr3bp,h,i,j,c,si,sj,sc,bound,pi1,pi2,implicitFunctionDerivative);
      }catch(...) {}

      // First we call acceptEnclosure. This method checks, if the bounds on curve from the previous step and the current step do overlap.
      // Otherwise we cannot guarantee, that two curves glue in a continuous way.
      if(oneStepResult and this->acceptEnclosure(bound,current)){
        boundLast = bound;
        current = next;
        this->validatedRange = interval(first,current);
        lastSuccessSizeI = trialSI;
        lastSuccessSizeJ = trialSJ;
        this->result = true;
        if(HaloContinuationTask::testComputation){
          std::cout << "no=" << this->taskNo 
                    << ", p=" << this->validatedRange
                    << ", si=" << trialSI.rightBound()
                    << ", sj=" << trialSJ.rightBound()
                    << ", step=" << this->stepParameter << std::endl;
          break;
        }

        // predict size of set and parameter range for the next step
        if(HaloContinuationTask::isInitialBoxSizeMaximal){
          intersection(si*successBoxScalefactor,SI,si);
          intersection(sj*successBoxScalefactor,SJ,sj);
        } else {
          si *= successBoxScalefactor;
          sj *= successBoxScalefactor;
        }
        this->stepParameter*=HaloContinuationTask::successParamSizeFactor;
      } else {
        if(this->result){
          this->stepParameter *= HaloContinuationTask::failureParamSizeFactor;
          si = lastSuccessSizeI;
          sj = lastSuccessSizeJ;
        } else {
          this->stepParameter /= 2.;
        }
      }
      PRINT(this->stepParameter);
    } while(current<last and this->stepParameter>HaloContinuationTask::minParameterStep);
  }

  this->report();
} // run

// #########################################################################
/**
 * This function checks if
 (1) for all x\in c there holds a \cap (b+x) \neq \emptyset.
  @param a,b,c - intervals
  @return true - if the condition (1) is validated and false otherwise.
  @remark - the function may return false even if (1) is satisfied.    
 */ 
bool checkOverlapingIntervals(interval a, interval b, interval c){
  // There are three cases which guarantee the assertion.

  // Case 1. right(a) belongs to all intervals b+x, x\in c.
  if ( (a.rightBound() >=  b.leftBound() + c) and (a.rightBound() <= b.rightBound()+c) ) return true;

  // Case 2. left(a) belongs to all intervals b+x, x\in c.
  if ( (a.leftBound() >=  b.leftBound() + c) and (a.leftBound() <= b.rightBound()+c) ) return true;

  // Case 3. b+c \subset a
  return subset(b+c,a);  
}

// #########################################################################

bool checkNonemptyIntersection(std::vector<BaseTask*>& tasks, int i, int j, int c){
  for(unsigned k=1;k<tasks.size();++k)
    if(!checkNonemptyIntersection(
        static_cast<HaloContinuationTask*>(tasks[k-1])->boundLast,
        static_cast<HaloContinuationTask*>(tasks[k])->boundFirst,
        i,j,c,
        static_cast<HaloContinuationTask*>(tasks[k])->boundFirst.getElement_x(c).leftBound()
      )) {
        std::cout << "Failed for k=" << k << std::endl;
        return false;
      }
  return true;
}

// ######################################################################
/**
 This algorithm checks if two affine sets overlap along one edge.
 */ 
bool checkNonemptyIntersection(const AffineSet& s1, const AffineSet& s2, int i , int j, int c, double z)
{
  if(
      not checkOverlapingIntervals(z,s1.getElement_r(c),s1.getElement_x(c)) or
      not checkOverlapingIntervals(z,s2.getElement_r(c),s2.getElement_x(c))
    ) {      
      cout << "Fatal error: parameter ranges do not overlap.\nThis should not happen by the construction of the algorithm.\n";
      cout << "One of the following conditions is not satisfied:\n";
      cout << z << "\\in " << s1.getElement_x(c) << "+" << s1.getElement_r(c) << endl;
      cout << z << "\\in " << s2.getElement_x(c) << "+" << s2.getElement_r(c) << endl;
      exit(0);
    }
  
  interval a = s2.getElement_x(i) - s1.getElement_x(i) + s2.getElement_B(i,c)*(z-s2.getElement_x(c)) - s1.getElement_B(i,c)*(z-s1.getElement_x(c));
  interval b = s2.getElement_x(j) - s1.getElement_x(j) + s2.getElement_B(j,c)*(z-s2.getElement_x(c)) - s1.getElement_B(j,c)*(z-s1.getElement_x(c));
  if(
      (checkOverlapingIntervals(s1.getElement_r(i),s2.getElement_r(i),a) or checkOverlapingIntervals(s2.getElement_r(i),s1.getElement_r(i),-a)) and
      (checkOverlapingIntervals(s1.getElement_r(j),s2.getElement_r(j),b) or checkOverlapingIntervals(s2.getElement_r(j),s1.getElement_r(j),-b))       
    ) return true;

  MESSAGE("Fatal error: sets do not overlap along continuation.\nWe cannot guarantee that the curve of periodic orbits is continuous.\n");
  PRINT(b);
  PRINT(s1.getElement_r(j));
  PRINT(s2.getElement_r(j));
  PRINT(s2.getElement_x(j) - s1.getElement_x(j));
  PRINT(s2.getElement_B(j,c)*(z-s2.getElement_x(c)));
  PRINT(s1.getElement_B(j,c)*(z-s1.getElement_x(c)));
  return false;
}

// #########################################################################
/**
 * This function checks if an interval vector v is a subset of affine set s.
 */ 
bool checkInclusion(const capd::IVector& v, const AffineSet& s){
  return subset(matrixAlgorithms::gauss(s.get_B(),v-s.get_x()),s.get_r());
}

