#include <algorithm>
#include <iostream>
#include "SymBreakingBifurcationHaloAlgs.h"
#include "auxil.h"
#include "Logger.h"

using namespace capd;
using namespace std;

SymBreakingBifurcationHaloAlgs::SymBreakingBifurcationHaloAlgs(int dim)
  : y(dim), x0(dim), r0(dim), r(dim), X(2), u0(2), N(2), K(2),
    flowD(dim,dim), D(dim,dim), C(dim,dim),
    flowH(dim,dim), H(dim,dim),
    J(dim,dim,3)
{
  C[0][0] = C[2][2] = C[4][4] = C[6][6] = 1;
}

void SymBreakingBifurcationHaloAlgs::init(const LDVector& h, const LDMatrix& DP, LDVector& t, const IVector& box, interval mu){
  mu.split(x0[6],r0[6]);
  std::copy(h.begin(),h.end(),x0.begin());
  r0[2] = box[2];
  r[0] = box[0]; r[2] = box[2]; r[4] = box[4]; r[6] = r0[6];
  X[0] = box[0]; X[1] = box[4];

  C[0][6] = t[0]; C[4][6] = t[4]; // dependence of bif point (x(mu),y'(mu)) wrt mu
  C[4][0] = -DP[3][0]/DP[3][4];   // dependence of Lyapunov orbit y'(x) wrt x
}

// #########################################################################
/**
  Here we fix value of z.
  We want to find set of zeroes of a function
    (x,dy) -> pi_{dx,dz} P_mu(x,0,z,dx=0,dy,dz=0)
  parameterized by (mu,z). It is assumed that mu is an interval of nonzero width
*/
bool SymBreakingBifurcationHaloAlgs::proveBranchBifHaloOrbits(ICR3BP& icr3bp){
  MESSAGE("==============> proveBranchBifHaloOrbits started");
  bool result = false;
  
  // compute value of function at centre of set and parameters
  PC1Rect2Set s0(PC1Rect2Set::C0BaseSet(x0,C,IVector(7)),PC1Rect2Set::C1BaseSet(C));
  y = icr3bp.expm(s0,flowD);
  D = icr3bp.expm.computeDP(y,flowD);
  u0[0] = y[3] + r[6]*D[3][6] + r[2]*D[3][2]; 
  u0[1] = D[5][2]; 

  // compute derivative of the map wrt to parameters, i.e. z and mu
  PC2Rect2Set s(PC2Rect2Set::C0BaseSet(x0,C,r0),PC2Rect2Set::C1BaseSet(C));
  y = icr3bp.exc2pm(s,flowD,flowH);
  icr3bp.exc2pm.computeDP(y,flowD,flowH,D,H);
  u0[0] += sqr(r[6])*H(3,6,6) + r[6]*r[2]*H(3,6,2) + sqr(r[2])*H(3,2,2);
  u0[1] += r[6]*H(5,6,2) + r[2]*H(5,2,2); 

  s = PC2Rect2Set(PC2Rect2Set::C0BaseSet(x0,C,r),PC2Rect2Set::C1BaseSet(C));
  y = icr3bp.exc2pm(s,flowD,flowH);
  icr3bp.exc2pm.computeDP(y,flowD,flowH,D,H);

  IMatrix M({{D[3][0],D[3][4]},{H(5,0,2),H(5,4,2)}});
  N = K = -intervalKrawczykOperator(M,u0,X);
  result = subset(K,X);

  try{
    N = -intervalNewtonOperator(M,u0);
    if(result)
      intersection(K,N,N);
    else
      result = subset(N,X);
  }catch(...){}
  
  MESSAGE("==============> proveBranchBifHaloOrbits finished");
  return result;
}

// #########################################################################

bool SymBreakingBifurcationHaloAlgs::checkConvexity(ICR3BP& icr3bp, interval x, interval dy){
  MESSAGE("==============> checkConvexity started");
  r[0] = x; r[2] = 0.; r[4] = dy;

  PC2Rect2Set s(x0,C,r);
  y = icr3bp.exc2pm(s,flowD,flowH);
  icr3bp.exc2pm.computeDP(y,flowD,flowH,D,H);
  IMatrix M({{D[3][0],D[3][4]},{H(5,0,2),H(5,4,2)}});
  IVector b({D[3][2],H(5,2,2)});
  PRINT(M);
  // first we compute tangent to bifurcation curve - should be zero
  IVector t = solveLinearSystem(M,-b);

  // then second order derivative of bifurcation curve
  interval mu = x0[6]+r[6];
  IVector bifPointBound(6,(x0+C*r).begin());
  PCnRect2Set s3(x0,C,r,3);
  icr3bp.excnpm(s3,J);
  J = icr3bp.excnpm.computeDP(J);
  
  interval b1 = H(3,0,0)*sqr(t[0]) + H(3,4,4)*sqr(t[1]) + H(3,0,4)*t[0]*t[1] + H(3,0,2)*t[0] + H(3,2,4)*t[1] + H(3,2,2);
  interval b2 = J(5,0,0,2)*sqr(t[0]) + J(5,2,4,4)*sqr(t[1]) + J(5,0,2,4)*t[0]*t[1] + J(5,0,2,2)*t[0] + J(5,2,2,4)*t[1] + J(5,2,2,2);
  IVector t2 = solveLinearSystem(M,IVector({-2.*b1,-2.*b2}));
  PRINT(t2);

  // compute derivative and Hessian of Jacobi integral at bif point
  IMatrix DJ(1,6);
  IHessian D2J(1,6);
  icr3bp.Jacobi.setParameter(0,mu);
  y = icr3bp.Jacobi(bifPointBound,DJ,D2J);
  // and compute first and second derivative of H(u(z))
  IVector T1({t[0],0.,1.,0.,t[1],0.});
  IVector T2({t2[0],0.,0.,0.,t2[1],0.});

  IMatrix D2H(6,6);
  for(int i=0;i<6;++i)
    for(int j=0;j<6;++j)
      D2H[i][j] = D2J(0,i,j);

  bool result = isSingular(DJ.row(0)*T1) and !isSingular(DJ.row(0)*T2 + T1*(D2H*T1));

  MESSAGE("==============> checkConvexity finished");
  return result;
}
