#include <algorithm>
#include "SymBreakingBifurcationLyapAlgs.h"
#include "auxil.h"
#include "Logger.h"

using namespace capd;

SymBreakingBifurcationLyapAlgs::SymBreakingBifurcationLyapAlgs(int dim)
  : v(dim), u0(4), u(dim), r0(dim), r(dim),
    D(dim,dim), C(dim,dim)
{
  C.setToIdentity();
}

// #########################################################################

void SymBreakingBifurcationLyapAlgs::init(const LDVector& h, const LDMatrix& DP, LDVector& t, const IVector& box, interval mu){
  mu.split(u[4],r0[4]);
  r[4] = r0[4];
  r0[0] = r[0] = box[0];
  r[3] = box[3];
  
  u0[0] = u[0] = h[0];
  u0[3] = u[3] = h[4];
  
  C[0][4] = t[0];
  C[3][4] = t[4];                // dependence of bif point (x(mu),y'(mu)) wrt mu
  C[3][0] = -DP[3][0]/DP[3][4];  // dependence of Lyapunov orbit y'(x) wrt x
}

// #########################################################################
/**
  This function proves the existence of a branch of planar Lyapunov orbits parameterized by (mu,x).
  For a fixed interval x and mu it solves equation
    proj_{dx} P(x,0,0,dy) = 0 
  for dy.
  @param [in] (x,dy) - an approximate Lyapunov orbit at the center of parameter set (mu,x)
  @param [in] C - coordinate system in which preform computation
  @param [in] box - box centered at zero. We will validate the existence of solution set in the affine set (x,0,0,dy) + C*box 
  @param [in] mu - mass parameter                                                        :
  @return - true if the family of orbits is validated
*/
bool SymBreakingBifurcationLyapAlgs::proveBranchLyapunovOrbits(IPCR3BP& pcr3bp){
  MESSAGE("==============> proveBranchLyapunovOrbits started");

  // first we estimate value at the center by mean value formula
  pcr3bp.setMu(u[4]);
  C0HOTripletonSet s00(u0);
  interval fx = pcr3bp.pm(s00)[2];

  C1HORect2Set s0(C1HORect2Set::C0BaseSet(u,C,r0),C1HORect2Set::C1BaseSet(C));
  v = pcr3bp.expm(s0,D);  
  D = pcr3bp.expm.computeDP(v,D);
  fx += D[2][0]*r[0] + D[2][4]*r[4];
  
  // Then we compute derivative of Poincare map at the entire set
  C1HORect2Set s(C1HORect2Set::C0BaseSet(u,C,r),C1HORect2Set::C1BaseSet(C));
  v = pcr3bp.expm(s,D);
  D = pcr3bp.expm.computeDP(v,D);

  // and we compute interval Newton operator
  N = - fx/D[2][3];
  bool result =  subsetInterior(N,r[3]);

  MESSAGE("==============> proveBranchLyapunovOrbits finished");
  return result;
}

// #########################################################################

bool SymBreakingBifurcationLyapAlgs::checkMonotonicity(IPCR3BP& pcr3bp, interval x, interval dy){
  MESSAGE("==============> checkMonotonicity started");
  
  pcr3bp.Jacobi.setParameter(0,u[4]+r[4]);
  IVector gradient = pcr3bp.Jacobi.derivative( IVector(4,(u + C*r).begin()) ).row(0);  
  IVector Cx({C[0][0],C[1][0],C[2][0],C[3][0]});
  IVector Cdy({C[0][3],C[1][3],C[2][3],C[3][3]});
  interval dydx = D[2][0]/D[2][3];

  // we check, if the energy is monotone along branch of Lyapunov orbits  
  bool result = !isSingular( gradient*Cx ) and !isSingular(gradient*(Cx - dydx*Cdy));

  MESSAGE("==============> checkMonotonicity finished");
  return result;
}
