#include <fstream>
#include "main.h"
#include "HaloContinuationTask.h"
using namespace capd;
const static IMatrix Id = IMatrix::Identity(2);

// #########################################################################
/**
  * This function finds an approximate Halo orbit fixing one of the variables (x,dy,z)
  * and varying the remaining two, which are given as parameters. 
  * Initial condition for a Halo orbit satisfies equation
      proj_{dx,dz} P(x,0,z,0,dy,0) = (0,0)
  * We solve this equation for either
  * x(z) and dy(z) or
  * dy(x) and z(x) or
  * x(dy) and z(dy).
  * @param v - a seed point for the Newton method
  * @param p - index of first variable
  * @param q - index of second variable
*/
template<class CR3BP, class V>
V findHaloOrbit(CR3BP& cr3bp, V v, int p, int q, int pi1, int pi2,double tol){
  MESSAGE("==============> findHaloOrbit started");
  typename CR3BP::Matrix D(6,6);
  int k;
  PRINT(v);
  for(k=0;k<30;++k){
    V y = cr3bp.pm(v,D);
    D = cr3bp.pm.computeDP(y,D);
    V u({y[pi1],y[pi2]}); 
    typename CR3BP::Matrix M({{D[pi1][p],D[pi1][q]},{D[pi2][p],D[pi2][q]}});
    u = solveLinearSystem(M,u);
    v[p] -= u[0];
    v[q] -= u[1];
    if(fabs(toDouble(u[0]))<tol and fabs(toDouble(u[1]))<tol) break;
  }
  PRINT(v);
  MESSAGE("==============> findHaloOrbit finished");
  return v;
}
template LDVector findHaloOrbit<>(DCR3BP&,LDVector,int,int,int,int,double);
template MpVector findHaloOrbit<>(MpCR3BP&,MpVector,int,int,int,int,double);

// #########################################################################
/**
 * This function finds an approximate bifurcation point of Halo family from Lyapunov family.
 * It solves the equation 
  (x,dy) -> pi_{dx,dz} P(x,0,z=0,dx=0,dy,dz=0) = (0,0)
 * for (x,dy). Near bifurcation point we expect this function has no isolated solution.
 * Therefore we factorize the above equation by z and then apply Newton method.
 @param v - seed point for the Newton method
 @return an approximate bifurcation point.
 @note The mass parameter must be set in the solver cr3bp before call this algorithm.
*/
template<class CR3BP, class V>
V findBifHaloOrbit(CR3BP& cr3bp, V v, typename CR3BP::Matrix* _D){
  MESSAGE("==============> findBifHaloOrbit started");
  typename CR3BP::Matrix flowD(6,6), D(6,6);
  typename CR3BP::Hessian flowH(6,6), H(6,6);
  v[2] = 0; // set z=0
  for(int i=0;i<20;++i){
    V y = cr3bp.pm(v,flowD,flowH);
    cr3bp.pm.computeDP(y,flowD,flowH,D,H);
    V u({y[3],D[5][2]}); 
    PRINT(u);
    typename CR3BP::Matrix M({{D[3][0],D[3][4]},{H(5,0,2),H(5,4,2)}});
    u = solveLinearSystem(M,u);
    v[0] -= u[0];
    v[4] -= u[1];
    if(fabs(toDouble(u[0]))<1e-12 and fabs(toDouble(u[1]))<1e-12) break;
  }
  PRINT(v);
  if(_D) *_D = D;
  MESSAGE("==============> findBifHaloOrbit finished");
  return v;
}

template LDVector findBifHaloOrbit<>(DCR3BP&,LDVector, LDMatrix*);
template MpVector findBifHaloOrbit<>(MpCR3BP&,MpVector, MpMatrix*);

// #########################################################################
/**
  Here we fix value of z.
  We want to find a zero of a function
  (x,dy) -> pi_{dx,dz} P(x,0,z,dx=0,dy,dz=0)
*/
bool proveBifHaloOrbit(ICR3BP& cr3bp, IVector x0, IVector& box, interval mu){
  MESSAGE("==============> proveBifHaloOrbit started");
  bool result = false;
  cr3bp.setMu(mu);
  const int dim = 6;
  IVector y(dim);
  IMatrix flowD(dim,dim), D(dim,dim);
  IHessian flowH(dim,dim), H(dim,dim);
  
  IVector X({box[0],box[4]}), u0(2);  
  IVector x = x0 + box;
  x0[2] += box[2];

  C2Rect2Set s(x);
  y = cr3bp.c2pm(s,flowD,flowH);
  cr3bp.c2pm.computeDP(y,flowD,flowH,D,H);
  IMatrix M({{D[3][0],D[3][4]},{H(5,0,2),H(5,4,2)}});
  IVector b({D[3][2],2.*H(5,2,2)});
  PRINT(M);

  C1HORect2Set s0(x0);
  y = cr3bp.pm(s0,flowD);
  D = cr3bp.pm.computeDP(y,flowD);
  u0[0] = y[3]; u0[1] = D[5][2]; 
  PRINT(u0);

  IVector N = -intervalNewtonOperator(M,u0);
  PRINT(N);
  PRINT(X);
  result = subset(N,X);

  if(result)
  {
    interval s = interval(-1,1)*1.1;
    intersection(N[0]*s,X[0],box[0]);
  }

  MESSAGE("==============> proveBifHaloOrbit finished");
  return result;
}

// #########################################################################
/***
 * Auxiliary function that computes approximate derivative of bifurcation point (x(mu),y'(mu))
 * with respect to the mass parameter just by means of symmetric difference quotient. 
 */
LDVector computeMuTangentVector(DCR3BP& cr3bp, LDVector h, interval mu){
  cr3bp.setMu(mu.leftBound());
  LDVector L = findBifHaloOrbit(cr3bp,h);
  cr3bp.setMu(mu.rightBound());
  LDVector R = findBifHaloOrbit(cr3bp,h);
  LDVector t = ((R-L)/(mu.rightBound()-mu.leftBound()));
  return t;
}

// #########################################################################
/**
  Here we fix value of z.
  We want to find set of zeroes of a function
    (x,dy) -> pi_{dx,dz} P_mu(x,0,z,dx=0,dy,dz=0)
  parameterized by (mu,z). It is assumed that mu is an interval of nonzero width
*/
bool proveBranchBifHaloOrbits(DCR3BP& cr3bp, ICR3BP& icr3bp, LDVector h, IMatrix C, IVector& box, interval mu){
  MESSAGE("==============> proveBranchBifHaloOrbits started");
  PRINT(mu);
  PRINT(diam(mu));
  bool result = false;
  interval m0, dmu;
  mu.split(m0,dmu);

  const int dim = 7;
  IVector y(dim);
  IMatrix flowD(dim,dim), D(dim,dim);
  IHessian flowH(dim,dim), H(dim,dim);

  IVector X({box[0],box[4]}), u0(2);  
  IVector x0(dim), r0(dim), r(dim);
  std::copy(h.begin(),h.end(),x0.begin()); x0[6] = m0;
  r0[2] = box[2]; r0[6] = dmu;
  r[0] = box[0]; r[2] = box[2]; r[4] = box[4]; r[6] = dmu;

  // compute value of function at centre of set and parameters
  PC1Rect2Set s0(PC1Rect2Set::C0BaseSet(x0,C,IVector(7)),PC1Rect2Set::C1BaseSet(C));
  y = icr3bp.expm(s0,flowD);
  D = icr3bp.expm.computeDP(y,flowD);
  u0[0] = y[3] + dmu*D[3][6] + box[2]*D[3][2]; 
  u0[1] = D[5][2]; 
  PRINT(u0);

  // compute derivative of the map wrt to parameters, i.e. z and mu
  PRINT(x0);
  PRINT(C);
  PRINT(r);
  PC2Rect2Set s(PC2Rect2Set::C0BaseSet(x0,C,r0),PC2Rect2Set::C1BaseSet(C));
  y = icr3bp.exc2pm(s,flowD,flowH);
  icr3bp.exc2pm.computeDP(y,flowD,flowH,D,H);
  u0[0] += sqr(dmu)*H(3,6,6) + dmu*box[2]*H(3,6,2) + sqr(box[2])*H(3,2,2);
  u0[1] += dmu*H(5,6,2) + box[2]*H(5,2,2); // BYLO 2.*
  PRINT(u0);  
  PRINT(D[3][6]);
  PRINT(D[3][2]);
  PRINT(H(5,6,2));
  PRINT(H(5,2,2));

  PRINT(r);
  s = PC2Rect2Set(PC2Rect2Set::C0BaseSet(x0,C,r),PC2Rect2Set::C1BaseSet(C));
  y = icr3bp.exc2pm(s,flowD,flowH);
  icr3bp.exc2pm.computeDP(y,flowD,flowH,D,H);
  IMatrix M({{D[3][0],D[3][4]},{H(5,0,2),H(5,4,2)}});
  PRINT(IVector({D[3][2],H(5,2,2)}));
  PRINT(M);

  IVector K = -intervalKrawczykOperator(M,u0,X);
  PRINT(K);
  PRINT(X);
  result = subset(K,X);

  if(!result) // try interval Newton
  {
    IVector N = -intervalNewtonOperator(M,u0);
    result = subset(N,X);
    PRINT(N);
    PRINT(X);
  }
    // Then we have to check if the curve of Halo orbits is convex at bifPointBound
  if(result){
    r[2] = 0; // z=0
    PC2Rect2Set s(x0,C,r);
    y = icr3bp.exc2pm(s,flowD,flowH);
    icr3bp.exc2pm.computeDP(y,flowD,flowH,D,H);
    IMatrix M({{D[3][0],D[3][4]},{H(5,0,2),H(5,4,2)}});
    IVector b({D[3][2],H(5,2,2)});
    PRINT(M);
    // first we compute tangent to bifurcation curve - should be zero
    IVector t = solveLinearSystem(M,-b);

    // then second order derivative 
    // J will be jet of Poincare map.
    IJet J(dim,dim,3);
    CnRect2Set s3(x0,C,r,3);
    icr3bp.excnpm(s3,J);
    J = icr3bp.excnpm.computeDP(J);
    
    interval b1 = H(3,0,0)*sqr(t[0]) + H(3,4,4)*sqr(t[1]) + H(3,0,4)*t[0]*t[1] + H(3,0,2)*t[0] + H(3,2,4)*t[1] + H(3,2,2);
    interval b2 = J(5,0,0,2)*sqr(t[0]) + J(5,2,4,4)*sqr(t[1]) + J(5,0,2,4)*t[0]*t[1] + J(5,0,2,2)*t[0] + J(5,2,2,4)*t[1] + J(5,2,2,2);
    IVector t2 = solveLinearSystem(M,IVector({-2.*b1,-2.*b2}));
    PRINT(t2);
    
    // compute derivative and Hessian of Jacobi integral at bif point
    IVector bifPointBound(6,(x0+C*r).begin());
    IMatrix DJ(1,6);
    IHessian D2J(1,6);
    icr3bp.Jacobi.setParameter(0,mu);
    y = icr3bp.Jacobi(bifPointBound,DJ,D2J);
    // and compute first and second derivative of H(u(z))
    IVector T1({t[0],0.,1.,0.,t[1],0.});
    IVector T2({t2[0],0.,0.,0.,t2[1],0.});
    
    cout << "Bound on symmetry breaking bifurcation point: " <<  bifPointBound << endl;
    cout << "Bound on Jacobi constant at bifurcation point: " << y << endl;
    cout << "T1=" << T1 << endl;
    cout << "T2=" << T2 << endl;
    cout << "G=" << DJ << endl;
    cout << "derivative=" << DJ.row(0)*T1 << endl;
    result = result and isSingular(DJ.row(0)*T1);
    IMatrix D2H(6,6);
    for(int i=0;i<6;++i)
      for(int j=0;j<6;++j)
        D2H[i][j] = D2J(0,i,j);
    cout << "hessian=" << DJ.row(0)*T2 + T1*(D2H*T1) << endl << endl;
    result = result and !isSingular(DJ.row(0)*T2 + T1*(D2H*T1));
  }
      
  MESSAGE("==============> proveBranchBifHaloOrbits finished");
  return result;
}

// #########################################################################
/**
  This function validates the existence of a branch of Halo orbits.
  We assume that u is an approximate Halo orbis and p,q are indices of variables
  that have to be solved with repsect to the parameter.

  The algorithm is based on a strighforward application of the Interval Newton Operator.
  It works well when the range of parameter is tiny or a point.

  @param x0 - approximate Halo orbit.
              One coordinate of x0 (different than p,q) can be an interval.
              As a result we obtain the existence of parameterized curve of Halo with respect to this variable.  
  @param p - index of first dependent variable
  @param q - index of second dependent variable
  @param sp,sq - sizes of box in (p,q) subsapce
  @param pi1,pi2 - indices of two-dimenional projection onto symmetric subspace
  @return true if the existence of Halo orbits is validated.
*/

bool proveHaloOrbit(ICR3BP& cr3bp, IVector x0, int p, int q, double sp, double sq, IVector& box, int pi1, int pi2){
  MESSAGE("==============> proveHaloOrbit started");
  IMatrix D(6,6);
  interval ip = sp*interval(-1,1);
  interval iq = sq*interval(-1,1);

  IVector x = x0;
  x[p] += ip;
  x[q] += iq;
  
  C1HORect2Set s(x);
  IVector y = cr3bp.pm(s,D);
  D = cr3bp.pm.computeDP(y,D);
  
  IMatrix M({{D[pi1][p],D[pi1][q]},{D[pi2][p],D[pi2][q]}});
  PRINT(M);

  C0HOTripletonSet s0(x0);
  y = cr3bp.pm(s0);
  IVector u0({y[pi1],y[pi2]}); 
  PRINT(u0);

  IVector X({ip,iq});
  // Interval Krawczyk Operator   
  IMatrix C = matrixInverse(midMatrix(M));
  IVector K = -C*u0 + (Id-C*M)*X;
  bool result = subset(K,X);
  if(result) {
    box[p] = K[0];
    box[q] = K[1];
  } else {
    // Interval Newton Operator   
    IVector N = -solveLinearSystem(M,u0);
    PRINT(N);
    PRINT(X);
    result = subset(N,X);
    if(result) {
      box[p] = N[0];
      box[q] = N[1];
    }
  }
  
  MESSAGE("==============> proveHaloOrbit finished");
  return result;
}

// #########################################################################
/**
  This function validates the existence of a branch of Halo orbits.
  We assume that u is an approximate Halo orbis and p,q are indices of variables t
  hat have to be solved with repsect to the parameter.

  The algorithm applies the Interval Newton/Krawczyk Operator to the function with affine coordiante change.

  @param s0 - centre of the set (approximate Halo orbit) 
  @param s - set centered at s0 at which we validate the existence of Halo orbits
  @param i - index of first dependent variable
  @param j - index of second dependent variable
  @param c - index of independent variable
  @param pi1,pi2 - indices of two-dimenional projection onto symmetric subspace
  @return true if the existence of Halo orbits is validated.
  @param [out] r - if scuccess result of interval Newton/Krawczyk operator 
  @param [out] Dfp - if scuccess and Dfp!=nullptr it will be assigned to derivative of fixed point curve wrt parameter.
*/
template<class PM, class C2PM, class C1Set, class C2Set, class Vector>
bool proveBranchHaloOrbits(PM& pm, C2PM& c2pm, C1Set& s0, C2Set& s, Vector& r, int i, int j, int c, int pi1, int pi2, Vector* Dfp){
  typedef typename PM::MatrixType Matrix;
  typedef typename PM::HessianType Hessian;
  Vector g00(2), g(2), y(6);
  Matrix flowD(6,6), D(6,6);
  Hessian flowH(6,6), H(6,6);
  pm.getSolver().setMaxStep(typename Matrix::ScalarType(1.));
  c2pm.getSolver().setMaxStep(typename Matrix::ScalarType(1.));

  // We will apply interval Krawczyk or Newton operator to a parameterized map
  // Step 1. Estimate value at the centre by first order Taylor approximation wrt to parameter.
  // Here we use higher precision arithmetic to compute value and derivative at the centre of set.
  y = pm(s0,D);
  D = pm.computeDP(y,D);
  g00[0] = y[pi1]; g00[1]=y[pi2]; 
  g[0] = g00[0] + r[c]*D[pi1][c];
  g[1] = g00[1] + r[c]*D[pi2][c];
  // take derivative of Poincare map at the center.
  // Later we will add bound from 2nd order derivatives.
  Matrix Dxf({{D[pi1][i],D[pi1][j]},{D[pi2][i],D[pi2][j]}});
  if(Dfp!=nullptr)
    *Dfp = Vector({D[pi1][c],D[pi2][c]});
  
  PRINT(g00);
  PRINT(g);
  PRINT(Dxf);
  // Step 2. Add error bounds to the value at centre from second order derivatives evaluated at entire set.
  y = c2pm(s,flowD,flowH);
  c2pm.computeDP(y,flowD,flowH,D,H);

  // add error bound to the value at centre from second order derivative wrt to parametr
  auto delta = sqr(r[c]);
  g[0] += H(pi1,c,c)*delta;
  g[1] += H(pi2,c,c)*delta;

  Dxf[0][0] += r[c]*H(pi1,i,c) + 2.*r[i]*H(pi1,i,i) + r[j]*H(pi1,i,j);
  Dxf[0][1] += r[c]*H(pi1,j,c) + r[i]*H(pi1,j,i) + 2.*r[j]*H(pi1,j,j);
  Dxf[1][0] += r[c]*H(pi2,i,c) + 2.*r[i]*H(pi2,i,i) + r[j]*H(pi2,i,j);
  Dxf[1][1] += r[c]*H(pi2,j,c) + r[i]*H(pi2,j,i) + 2.*r[j]*H(pi2,j,j);

  PRINT(g);
  PRINT(Dxf);
  
  Vector X({r[i],r[j]});
  Vector K = -intervalKrawczykOperator(Dxf,g,X);
  PRINT(K);  
  PRINT(X);
  bool result = false;
  if(subset(K,X)){
    r[i] = K[0]; r[j] = K[1];
    result = true;
  }
  // try also Interval Newton Operator. It can improve bound but the matrix can be not invertible
  try{
    Vector N = - solveLinearSystem(Dxf,g);
    PRINT(N);  
    PRINT(X);
    if(subset(N,X)){
      intersection(r[i],N[0],r[i]);
      intersection(r[j],N[1],r[j]);
      result = true;
    }
  } catch(...) {}
  if(result and Dfp!=nullptr){
    (*Dfp)[0] += 2.*r[c]*H(pi1,c,c) + r[i]*H(pi1,c,i) + r[j]*H(pi1,c,j);
    (*Dfp)[1] += 2.*r[c]*H(pi2,c,c) + r[i]*H(pi2,c,i) + r[j]*H(pi2,c,j);

    // Dfp is derivative of implict function z->(x(z),dy(z)) which is curve of fixed points
    *Dfp = -solveLinearSystem(Dxf,*Dfp);    
  }
  return result;
}

template
bool proveBranchHaloOrbits<>(
    MpICR3BP::MpPoincareMap&, MpICR3BP::MpPoincareMap&,
    MpC1Rect2Set&, MpC2Rect2Set&,
    MpIVector& ,
    int,int,int,int,int,
    MpIVector*
  );
// #########################################################################
/**
 * Auxiliary function. It computes (nonrigorously) tangent vector
 * to implicit curve at a given point.
*/
LDVector computeTangentVector(DCR3BP& dcr3bp, LDVector v, int i, int j, int c, int pi1, int pi2){
  LDMatrix Df(6,6);
  v = dcr3bp.pm(v,Df);
  Df = dcr3bp.pm.computeDP(v,Df);
  
  LDVector Dz0f({Df[pi1][c],Df[pi2][c]});
  LDMatrix Dx0f({{Df[pi1][i],Df[pi1][j]},{Df[pi2][i],Df[pi2][j]}});

  return solveLinearSystem(Dx0f,Dz0f); 
}

// #########################################################################
/**
  This function validates the existence of a branch of Halo orbits.
  We assume that u is an approximate Halo orbis and p,q are indices of variables t
  hat have to be solved with repsect to the parameter.

  The algorithm applies the Interval Newton Operator to the function with affine coordiante change.

  @param v - approximate Halo orbit
  @param i - index of first dependent variable
  @param j - index of second dependent variable
  @param c - index of independent variable
  @param si,sj - box in (i,j) subsapce
  @param sc - interval of parameters (range of c-coordinate)
  @param pi1,pi2 - indices of two-dimenional projection onto symmetric subspace
  @return true if the existence of Halo orbits is validated.
*/
bool proveBranchHaloOrbits(DCR3BP& dcr3bp, ICR3BP& cr3bp, LDVector v, int i, int j, int c, capd::interval& si, capd::interval& sj, interval sc, AffineSet& bound, int pi1, int pi2, IVector* Dfp){
  MESSAGE("==============> proveBranchHaloOrbits started");
  // first we compute change of coordinates (nonrigorously)
  IVector A = vectalg::convertObject<IVector>(computeTangentVector(dcr3bp,v,i,j,c,pi1,pi2)); 
  PRINT(A);
  PRINT(sc);

  IMatrix D = IMatrix::Identity(6), M(6,6);
  M[i][c] = D[i][c] = -A[0];
  M[j][c] = D[j][c] = -A[1];
  M[i][i] = M[j][j] = M[c][c] = 1;
  // setting three sets for integration
  IVector x(6), r(6);
  x[i] = v[i];
  x[j] = v[j];
  r[i] = si;
  r[j] = sj;
  sc.split(x[c],r[c]);
  bound = AffineSet(x,D,r);

  typedef C1HORect2Set C1Set;
  C1Set s0(C1Set::C0BaseSet(x,M,IVector(6)),C1Set::C1BaseSet(M));
  C2Rect2Set s(C2Rect2Set::C0BaseSet(x,M,r),C2Rect2Set::C1BaseSet(M));     // entire affine bound 
  
  bool result = proveBranchHaloOrbits(cr3bp.pm,cr3bp.c2pm,s0,s,r,i,j,c,pi1,pi2,Dfp);
  if(result){
    si = r[i];
    sj = r[j];
  }
  MESSAGE("==============> proveBranchHaloOrbits finished");
  return result;
}

