/**
  This file contains implementation on routines regarding Lyapunov orbits.
  They 
  - find approximate planar and vertical Lyapunov orbits
  - prove the existence of the family of planar Lyapunov orbits
    within a given set of velocities and for a specified range of mu. 
*/

#include <stdexcept>
#include "main.h"
using namespace capd;

// #########################################################################
/**
  This function proves the existence of planar Lyapunov orbit.
  For a fixed interval of dy it solves equation
    proj_{dx} P(x,0,0,dy) = 0 
  for x.
  @param [in] u0 - an approximate bifurcation point 
  @param [in] box - a box around x0 in which we want to prove the Lyapunov family
  @param [in] mu - mass parameter
  @param [out] box - if result=true, it is shrinked so that u0+box it still contains enclosure of  the curve dy->x(dy)
  @return - true if the family of orbits is validated
*/
bool proveLyapunovOrbit(IPCR3BP& pcr3bp, IVector u0, IVector& box, interval mu){
  MESSAGE("==============> proveLyapunovOrbit started");
  pcr3bp.setMu(mu);

  IVector u = u0+box;
  // bound on gradient of Jacobi integral on the domain 'u'.
  IVector g  = pcr3bp.Jacobi.derivative(u).row(0);
  u0[3] += box[3];
  C0HOTripletonSet s0(u0);
  C1HORect2Set s(u);
  IMatrix D(4,4);
  IVector y = pcr3bp.pm(s,D);
  D = pcr3bp.pm.computeDP(y,D);
  IVector y0 = pcr3bp.pm(s0);

  interval N = - y0[2]/D[2][0];
  PRINT(y0);
  PRINT(N);
  PRINT(box[0]);
  PRINT(g[3]);
  PRINT(g[3] - g[0]*D[2][3]/D[2][0]);
  // check the inclusion for Newton operator and the condition for monotonicity of energy, i.e.
  // D_{dy}C\neq 0 and D_{dy}C + D_{x}C *x'(dy) \neq 0 
  bool result = subsetInterior(N,box[0]) and !isSingular(g[3]) and !isSingular(g[3] - g[0]*D[2][3]/D[2][0]);

  if(result){
    intersection(N*interval(-1,1)*1.1,box[0],box[0]);
  }

  MESSAGE("==============> proveLyapunovOrbit finished");
  return result;
}

// #########################################################################
/**
  This function proves the existence of a branch of planar Lyapunov orbits parameterized by (mu,x).
  For a fixed interval x and mu it solves equation
    proj_{dx} P(x,0,0,dy) = 0 
  for dy.
  @param [in] (x,dy) - an approximate Lyapunov orbit at the center of parameter set (mu,x)
  @param [in] C - coordinate system in which preform computation
  @param [in] box - box centered at zero. We will validate the existence of solution set in the affine set (x,0,0,dy) + C*box 
  @param [in] mu - mass parameter
  @return - true if the family of orbits is validated
*/
bool proveBranchLyapunovOrbits(IPCR3BP& pcr3bp, double x, double dy, IMatrix C, IVector& box, interval mu){
  MESSAGE("==============> proveBranchLyapunovOrbits started");
  interval m0, dmu;
  mu.split(m0,dmu);
  IMatrix D(5,5);
  IVector v(5);
  
  IVector u0({x,0.,0.,dy});
  IVector u({x,0.,0.,dy,m0});
  IVector r0({box[0],0.,0.,0.,dmu});
  IVector r({box[0],0.,0.,box[3],dmu});

  // first we estimate value at the center by mean value formula
  pcr3bp.setMu(m0);
  C0HOTripletonSet s00(u0);
  IVector y00 = pcr3bp.pm(s00);
  C1HORect2Set s0(C1HORect2Set::C0BaseSet(u,C,r0),C1HORect2Set::C1BaseSet(C));
  v = pcr3bp.expm(s0,D);  
  D = pcr3bp.expm.computeDP(v,D);

  interval fx = y00[2] + D[2][0]*box[0] + D[2][4]*dmu;
  PRINT(fx);
  PRINT(y00[2]);
  PRINT(v[2]);
  PRINT(D[2][0]);
  PRINT(D[2][4]);
  
  // Then we compute derivative of Poincare map at the entire set
  C1HORect2Set s(C1HORect2Set::C0BaseSet(u,C,r),C1HORect2Set::C1BaseSet(C));
  v = pcr3bp.expm(s,D);
  D = pcr3bp.expm.computeDP(v,D);

  // and we compute interval Newton operator
  interval N = - fx/D[2][3];
  PRINT(N);
  PRINT(box[3]);
  bool result =  subsetInterior(N,box[3]);

  // Finally, we check if the Hamiltonian DH/dx is nonzero and d/dx H(dy(x)) is nonzero
  IVector gradient = pcr3bp.Jacobi.derivative( IVector(4,(u + C*r).begin()) ).row(0);  
  IVector Cx({C[0][0],C[1][0],C[2][0],C[3][0]});
  IVector Cdy({C[0][3],C[1][3],C[2][3],C[3][3]});
  result = result and !isSingular( gradient*Cx );
  interval dydx = D[2][0]/D[2][3];
  // we check, if the energy is monotone along branch of Lyapunov orbits  
  result = result and !isSingular(gradient*(Cx - dydx*Cdy));
  MESSAGE("==============> proveBranchLyapunovOrbits finished");
  return result;
}
